# -*- coding: utf-8 -*-

######################################
#
#   Butler-Addon für Anki
#
#   Copyright: Thomas Kahn (veröffentlicht 26.02.2020)
#   Lizenz:    AGPL, Version 3, https://www.gnu.org/licenses/agpl-3.0.html
#   Hinweis 1: Diese Lizenz bezieht sich ausschließlich auf den Inhalt dieser Datei. Die Jura-Vorlagen oder Basiskarten sind davon NICHT umfasst!
#   Hinweis 2: Bei Änderungen am Code muss ggf. der Datenschutz-Hinweis angepasst werden, siehe die Funktion namens "datenschutz_info()".
#
#
#   Credits:
#
#   htmldiff.js (wird verwendet, um die Veränderungen in einem Update anzuzeigen)
#   Autor:   https://github.com/tnwinc
#   Github:  https://github.com/tnwinc/htmldiff.js
#   Lizenz:  MIT License
#
###################################### 

import time

# Regular Expressions
import re

# import the main window object (mw) from aqt
from aqt import mw, browser, gui_hooks, dialogs
# import the "show info" tool from utils.py
from aqt.utils import showInfo, showText, askUser, askUserDialog, tooltip, getText, chooseList
# import all of the Qt GUI library
from aqt.qt import *

# Für Umgang mit Pfaden
from anki.utils import isMac, isWin, isLin

from anki.lang import _ #Für die Zukunft wichtig laut anki-console.exe

# Für HTTP Requests und Kommunikation mit dem Server (Abgleich, ob eine neue Version erschienen ist)
import json
import datetime
import urllib

#Um darauf zu warten, dass ein Profil in Anki geladen wurde
from anki.hooks import addHook

#Um Dateien zu öffnen
import shlex

#Version 2.0 oder 2.1+?
from anki import version
anki20 = version.startswith("2.0.")
if anki20:
    import httplib2
else:
    from urllib.request import urlretrieve, urlopen
    import requests
    from anki.sync import AnkiRequestsClient
    
#Zum Extrahieren der Addons
from io import BytesIO
from zipfile import ZipFile
import os



######################################
#
#   EINSTELLUNGEN UND LINKS
#
######################################   

#Einstellungen für das Butler-Addon:
first_start                    = True
update_interval                = 60 * 24 #min
butler_addon_version_lokal     = 36
butler_addon_automatic_updates = True

#Links für das Butler-Addon:
butler_addon_version_remote = u"http://thomaskahn.de/anki-butler-addon/updates/addon_version.txt"
butler_addon_settings_info  = u"http://thomaskahn.de/anki-butler-addon/#/settings"
butler_addon_dl             = u"http://thomaskahn.de/anki-butler-addon/updates/butler.py"
butler_addon_changelog      = u"http://thomaskahn.de/anki-butler-addon/#/news"
butler_addon_latest_info    = u"http://thomaskahn.de/anki-butler-addon/API/get_latest_all.php"

#Einstellungen und Links für die empfohlenen Addons:
recommended_plugins_stand_lokal       = "2000-01-01 00:00"
recommended_plugins_info              = u"http://thomaskahn.de/anki-butler-addon/#/addons"
recommended_plugins_stand_remote_21   = u"http://thomaskahn.de/anki-butler-addon/API/get_current_addons_version.php"
recommended_plugins_21                = u"http://thomaskahn.de/anki-butler-addon/API/dl_addons.php"
recommended_plugins_release_notes     = u"http://thomaskahn.de/anki-butler-addon/API/get_addons_release_notes.php"
recommended_plugins_changelog         = u"http://thomaskahn.de/anki-butler-addon/#/news"
recommended_plugins_automatic_updates = True

#Einstellungen und Links zum Updaten der Basiskarten:
bk_stand       = u"http://thomaskahn.de/basiskarten/current_state_of_decks.php"
bk_update_link = u"http://thomaskahn.de/basiskarten/update.php?upd_id="
bk_dl_updates_link = u"http://thomaskahn.de/basiskarten/get_updates_for_deck.php"
bk_nids_url    = u"http://thomaskahn.de/basiskarten/nids.txt"   
bk_cids_and_nids_url = u"http://thomaskahn.de/basiskarten/cids_and_nids.txt"   
bk_automatic_updates = True
bk_tagged      = False
bk_tagged2     = False
token = u""
vv_hinweis_anzeigen = True

#Einstellungen und Links für die Jura-Vorlagen:
jura_vorlagen_stand_remote      = u"http://thomaskahn.de/anki-butler-addon/API/get_current_juvo_version.php"
jura_vorlagen_dl                = u"http://thomaskahn.de/anki-butler-addon/API/dl_juvo_vollinstallation.php"
jura_vorlagen_content           = u"http://thomaskahn.de/anki-butler-addon/API/dl_juvo_update.php"
jura_vorlagen_zusatz            = u"http://thomaskahn.de/anki-butler-addon/API/dl_juvo_zusaetzliche_dateien.php"
jura_vorlagen_changelog         = u"http://thomaskahn.de/jura-vorlagen/#/news"
jura_vorlagen_release_notes     = u"http://thomaskahn.de/anki-butler-addon/API/get_jura_vorlagen_release_notes.php"
jura_vorlagen_automatic_updates = True

#Teilnahme_am_Beta-Programm
beta_asked       = False
beta_participate = False
beta_email       = u""

#Nachrichten
nachrichten_stand_remote = u"http://thomaskahn.de/anki-butler-addon/API/get_latest_news_date.php"
nachrichten_url          = u"http://thomaskahn.de/anki-butler-addon/API/get_latest_news_inhalt.php"
nachrichten_stand_lokal  = "2000-01-01 00:00"
nachrichten_empfangen    = True

if isWin:
    pfad_trenner = "\\"
else:
    pfad_trenner = "/"

######################################
#
#   ALLGEMEINE FUNKTIONEN
#
######################################

def open_file(filename):
    if isWin:
        os.startfile(filename) #Zeigt kein hässliches Kommandozeilen-Fenster! 
        #os.system("start " + 'jura-vorlagen.apkg') #Windows only
    elif isMac:
        os.system("open " + shlex.quote(filename)) #Funktioniert auf Mac, aber auf Linux nicht.
    else:
        showInfo(u"Bitte manuell öffnen: " +filename)
    
def http_get(url):
    f = urlopen(url)
    return f.read().decode()
    
    # Alte Variante:
    # try:
        # if anki20:
            # resp, content = httplib2.Http().request(url);
            # return content
        # else:
            # f = urlopen(url)
            # return f.read().decode()
    # except:
        # pass
        
def http_post(url, json_data):
    r = requests.post(url, data=json_data)
    return(r.text)

    data = urlencode(json_data, doseq=True)
    req  = Request(url, data)
    response = urllib2.urlopen(req)
    return response.read()

def dl_file_to(url, target):
    try:
        # Alte Variante:
        # if anki20:
            # urllib.urlretrieve(url, target)
        # else:
            # urlretrieve(url, target)
        urlretrieve(url, target)
    except:
        #showInfo(u"Keine Verbindung zum Server möglich! (Stelle: dl_file_to)")
        return False
    return True
        
def dl_file_and_unzip_to(url, target):
    try:
        request = requests.get(url)
    except:
# Disabled since Anki 2.1.29, see: https://forums.ankiweb.net/t/anki-2-1-29-win-mac-attributeerror-nonetype-object-has-no-attribute-cancel/1598        mw.progress.finish()
        showInfo(u"Fehler: Datei konnte nicht heruntergeladen werden! (Stelle: dl_file_and_unzip_to #1)")
        return False
    try:
        zip = ZipFile(BytesIO(request.content))
        zip.extractall(target)
    except:
# Disabled since Anki 2.1.29, see: https://forums.ankiweb.net/t/anki-2-1-29-win-mac-attributeerror-nonetype-object-has-no-attribute-cancel/1598        mw.progress.finish()
        showInfo(u"Fehler: Datei konnte nicht entpackt werden! (Stelle: dl_file_and_unzip_to #2)")
        return False
    return True



######################################
#
#   NÜTZLICHES
#
######################################

def configure_settings(silent = False):
    if silent or askUser(u"Welche Einstellungen durch diese Funktion verändert werden, <a href=\"" +butler_addon_settings_info +u"\">kannst du hier nachlesen</a>.<br><br>Jetzt Einstellungen vornehmen?"):
      try:
        mw.col.set_config("activeCols", [u'noteCrt', u'cardDue', u'noteFld', u'template', u'noteMod'])
        mw.col.set_config("sortType", u'noteCrt')
        mw.col.set_config("sortBackwards", False)
        mw.col.set_config("activeNoteCols", [u'noteCrt', u'noteFld', u'note', u'noteMod'])
        mw.col.set_config("noteSortType", u'noteCrt')
        mw.col.set_config("browserNoteSortBackwards", False)
        mw.col.set_config("newSpread", 1)        
      except:
        mw.col.conf["activeCols"] = [u'noteCrt', u'cardDue', u'noteFld', u'template', u'noteMod']
        mw.col.conf["sortType"] = u'noteCrt'
        mw.col.conf["sortBackwards"] = False
        try:
          mw.col.conf["activeNoteCols"] = [u'noteCrt', u'noteFld', u'note', u'noteMod']
          mw.col.conf["noteSortType"] = u'noteCrt'
          mw.col.conf["browserNoteSortBackwards"] = False
        except:
          pass
        mw.col.conf["newSpread"] = 1 # -- NEW_CARDS_LAST (see new cards after review)
        #mw.col.save()  #Tut nichts, glaube ich.
        mw.col.setMod() #Thx, glutanimate! https://anki.tenderapp.com/discussions/add-ons/8512-cannot-save-addon-configuration-in-mwcolconf-gets-removed-on-sync
      mw.pm.profile['hideDeckLotsMsg'] = True
      mw.pm.save()
      mw.reset()

        # TO DO Breite der Kategorien anpassen
      if not silent:
          showInfo(u"Danke, das war's. Anki wurde soeben optimal für akademisches Wissen eingestellt. Du musst das in Zukunft nicht nochmal tun - es sei denn, du hast etwas verstellt und möchtest die Einstellungen zurücksetzen.")
            
def open_media_folder():
    open_file(os.path.join(mw.pm.profileFolder(), "collection.media"))

def open_addon_folder():
    pfad = os.path.dirname(os.path.realpath(__file__))
    if isWin:
        plugin_ordner = pfad[:pfad.rindex("\\")]
    else:
        plugin_ordner = pfad[:pfad.rindex("/")]
    open_file(plugin_ordner)
    
def open_backup_folder():
    open_file(mw.pm.backupFolder())

def open_update_log_folder():
    open_file(mw.pm.profileFolder())

def find_notes(terms):
    try:
      return mw.col.find_notes(terms)
    except:
      return mw.col.findNotes(terms)
      
def find_cards(terms):
    try:
      return mw.col.find_cards(terms)
    except:
      return mw.col.findCards(terms)

def forget_cards(list_of_cids):
    if list_of_cids == []:
      return
    mw.col.sched.schedule_cards_as_new(list_of_cids)
    mw.col.sched.reset()

def lookup_browser(text):
  #From: https://ankiweb.net/shared/info/930944997
  browser = dialogs.open("Browser", aqt.mw)
  browser.form.searchEdit.lineEdit().setText(text)
  browser.onSearchActivated()



######################################
#
#   BASISKARTEN
#
######################################  

#Deckname aus Deck-Description filtern
def deck_id_in_desc(desc):
    try:
        return desc.split(u"basiskarten.de/")[1].split(u">Hier")[0].strip("\"").replace('zpo_i','zpo1').replace('_', '') #Nötig, weil ich irgendwann mal den Feedback-Link von "zpo_i" in "zpo1" geändert habe. Das zweite replace ist nur zur Sicherheit da, falls ich das nochmal getan haben sollte.
    except:
        return ""
    #return desc.split("basiskarten.de")[0].split(">Hier")[0]#.trim("\"")

#Aktuellen Stand eines Stapels aus der Stapel-Beschreibung auslesen. Nur auf BK-Stapel anwenden, sonst macht das Ergebnis keinen Sinn.

def bk_id_to_day(bk_id):
  return datetime.datetime.fromtimestamp(round(bk_id / 1000)).strftime("%Y-%m-%d")

def get_deck_date_from_desc(deck, bk_name):
  try:  
    if deck_id_in_desc(deck["desc"]) == bk_name:
      if "<span id=basiskarten_stand>" in deck["desc"]:
        m = re.search('<span id=basiskarten_stand>(\d+-\d+-\d+)+<\/span>', deck["desc"])
        return m.group(1) #Scheitert mit Fehlermeldung, wenn die Gruppe nicht gefunden wird.
    else:
      return -1
  except:
    return -1
  return "1909-02-26" #http://www.deutschestextarchiv.de/book/show/ebbinghaus_gedaechtnis_1885

def get_deck_version_from_desc(deck, bk_name):
  try:
    if deck_id_in_desc(deck["desc"]) == bk_name:
      if "<span id=basiskarten_version>" in deck["desc"]:      
        m = re.search('<span id=basiskarten_version>(\d+)<\/span>', deck["desc"])
        return int(m.group(1)) #Scheitert mit Fehlermeldung, wenn die Gruppe nicht gefunden wird.
    else:
      return -1
  except:
    return -1
  return 0

def get_deck_version_from_cards_in_deck(deck, bk_name, acceptable_dates = []):
  update = 0
  if deck_id_in_desc(deck["desc"]) == bk_name:
    while True:
      found_nids = find_notes("\"deck:{}\" \"[basiskarten_update\: {}]\"".format(deck["name"], update +1))
      updated_notes_of_this_deck = []
      for nid in found_nids:
        bk_id = get_basiskarten_id(mw.col.getNote(nid))
        if bk_id != None and ( acceptable_dates == [] or bk_id_to_day(bk_id) in acceptable_dates ):
          updated_notes_of_this_deck.append(nid)
      
      if len(updated_notes_of_this_deck) == 0:
        break
      update += 1
  else:
    return -1
  return update
  
  #✅ Problem: Funktioniert nicht, wenn jemand andere Basiskarten mit höherer Update-Nummer in den Stapel geschoben hat.
  #❗️ Zusätzlich akzeptierte Daten für jeden Stapel angeben? Ja, muss so sein.  
 
  print(u"=> Höchste basiskarten_update-Version im Stapel {}: {}".format(bk_name, str(update))) #deck["name"]
  # print("")
  # print("")
  
def fillstr(str):
  while len(str) < 13:
    str += "_"
  return str

def get_deck_version_from_cards_only(bk_id_ranges, current_version, min_bk_notes_count):
  if bk_id_ranges == []:
    return -1

  suchstr = ""
  for ids in bk_id_ranges:
    suchstr += " OR \"Quellen:*[basiskarten_id: " +fillstr(ids) +"]*\""
  suchstr = "(" +suchstr.strip(" OR ") +")"
  
  #print(str(suchstr +" tag:Basiskarten") +' Line: ' +str(sys._getframe().f_lineno))
  found_nids = find_notes(suchstr)
  # print(str(len(found_nids)) +' Line: ' +str(sys._getframe().f_lineno) +"\n")

  if len(found_nids) < min_bk_notes_count[0]:
    return -1 #Keine oder nicht genügend Notizen (!) aus diesem Stapel gefunden.

  update = current_version
  while update >= 1:
    found_nids = find_notes(suchstr +" \"Quellen:*[basiskarten_update: " +str(update)+ "]*\"")
    if len(found_nids) >= min_bk_notes_count[update]:
      break
    update -= 1
  return update

#Hat der User überhaupt Basiskarten-Stapel?
def user_has_basiskarten():
  return len(find_notes("tag:Basiskarten")) > 0

#Welche Basiskarten-Stapel hat der User?
def search_for_known_decks(inverse = False):

  if inverse:
    found_decks = []
  else:
    found_decks = {} # NEU:
    # { 
      #"version" : Höchste gefundene Version,
      #"dids"    : Array aller dids mit dieser Stapel-ID,
      #"differing_versions" : Unterschiedliche Versionen in Stapeln gefunden ja/nein
    # }

    # ALT: { bk_name : ( [0] Datum , [1] Name des Subdecks , [2] minimale Länge des vollen Decknamens, [3] deck-id ) }
  
  #Find best way to loop through decks
  old_way = False
  try:
    all_decks = mw.col.decks.all_names_and_ids()
    #print('Butler-Addon: Using new way to access decks. Line: ' +str( sys._getframe().f_lineno))
  except:
    #print('Butler-Addon: Using OLD way to access decks. Line: ' +str( sys._getframe().f_lineno))
    
    old_way = True
    all_decks = mw.col.decks.decks.items()

  for deck in all_decks: #Alle Stapel durchgehen
    if not old_way:
      did = str(deck.id)
      deck_json = mw.col.decks.get(deck.id)
    else:
      did = deck[0]
      deck_json = mw.col.decks.decks[did]
    
    bk_name = deck_id_in_desc(deck_json["desc"])
    
    # if deck_json['desc'] == "":
      # print(deck_json['name'] +" => EMPTY: ")
    # else:
      # print(deck_json['name'] +" => "+bk_name)
       
    if inverse:
      condition = bk_name == ""
    else:
      condition = bk_name != ""
       
    if condition: #Wir suchen nur nach Basiskarten-Decks.
      if not old_way:
        deck_name = deck.name
      else:
        deck_name = deck_json["name"]  

      if deck_name.find("::") == -1:
        versatz = 1
      else:
        versatz = 2        
      
      if inverse:
        found_decks.append( {
          "name" : deck_name[deck_name.rfind("::") +versatz:],
          "path" : deck_name,
          "did"  : did,
          "deck_obj" : deck_json
        } )
        
      else:
        if not bk_name in found_decks:
          found_decks[bk_name] = {
            "date" : get_deck_date_from_desc(deck_json, bk_name),
            "version" : get_deck_version_from_desc(deck_json, bk_name),
            #"dids"  : [did],
            "differing_versions" : False
          }
        else:
          new_version = get_deck_version_from_desc(deck_json, bk_name)
          new_date = get_deck_date_from_desc(deck_json, bk_name)
          
          if new_version != found_decks[bk_name]["version"] or new_date != found_decks[bk_name]["date"]:
            found_decks[bk_name]["differing_versions"] = True

          if new_version > found_decks[bk_name]["version"]:
            found_decks[bk_name]["version"] = new_version
          
          if new_date > found_decks[bk_name]["date"]:
            found_decks[bk_name]["date"] = new_date
          
          #found_decks[bk_name]["dids"].append(did)
          
        # if not bk_name in found_decks or len(deck_name) < found_decks[bk_name][2]: #Hauptstapel mit diesem bk_name finden. Der Hauptstapel ist der Stapel des jeweiligen bk_name mit dem kürzesten Stapelnamen.
          # add = { bk_name : (get_deck_date_from_desc(deck_json, bk_name), deck_name[deck_name.rfind("::") +versatz:], len(deck_name), did) }
          # found_decks.update(add)
  
  if not inverse:
    print(u"Butler-Addon: {} Basiskarten-decks found: {}".format(len(found_decks), ", ".join(found_decks.keys())))
  else:
    print(u"Butler-Addon: {} Non-Basiskarten-decks found.".format(len(found_decks)))
  
  # print('Butler-Addon (Line: ' +str(sys._getframe().f_lineno) +"):\n" +json.dumps(found_decks, indent = 4))
  
  return found_decks
  
  
  
def check_for_deck_updates(info_all = None, notify_of_result = True):
  
  #Prüfen, ob überhaupt Basiskarten in Sammlung
  if not user_has_basiskarten():
    if notify_of_result:
      showInfo(u"<h3>Es wurden keine Basiskarten in deiner Sammlung gefunden.</h3>", title="Butler-Addon")
    return False
  
  print("Checking if Basiskarten-decks are up to date (Butler-Addon).")
  
  #Stand der Stapel auf dem Server abfragen
  if info_all == None:
    info_all = get_info(notify_of_result = notify_of_result)
  if info_all == None:
    return False
  
  decksRemote = info_all['bk']

  decks_outdated_str   = u""
  decks_outdated       = {}
  decks_uptodate_str   = u""
  decks_uptodate_count = 0
  cat_outdated = {}
  cat_uptodate = {}
  
  stand_aller_bk_stapel_laut_stapelbeschreibung = search_for_known_decks()
  for curr, val in decksRemote.items():
    
    # print(val[2]) #Stapelname
    # print(val[4])
    # print(val[3])
    
    stand_laut_karten = get_deck_version_from_cards_only(val[4], val[3], val[6])

    stand_laut_stapelbeschreibung = -1
    if curr in stand_aller_bk_stapel_laut_stapelbeschreibung:
      stand_laut_stapelbeschreibung = stand_aller_bk_stapel_laut_stapelbeschreibung[curr]["version"]
      
      #Stapelbeschreibungen angleichen, falls unterschiedliche Versionen vorhanden
      if stand_aller_bk_stapel_laut_stapelbeschreibung[curr]["differing_versions"] == True:
        save_deck_status(curr, stand_aller_bk_stapel_laut_stapelbeschreibung[curr]["version"], stand_aller_bk_stapel_laut_stapelbeschreibung[curr]["date"])
        print("===\nStapelbeschreibung von {} korrigiert\n".format(curr))
   
    if stand_laut_karten == -1:
      #Vermutlich nur Probierstapel
      stand_lokal = -1
    elif stand_laut_karten == stand_laut_stapelbeschreibung:
      stand_lokal = stand_laut_karten
    else:
      stand_lokal = stand_laut_karten
      save_deck_status(curr, stand_laut_karten, "")
      print("===\nStapelbeschreibung von {} korrigiert\n".format(curr))
    
    # DEBUG: Welche Version haben die verschiedenen Basiskarten-Stapel in der Sammlung?
    # print(val[2]) #Stapelname
    # print("stand_laut_karten:             " +str(stand_laut_karten))
    # print("stand_laut_stapelbeschreibung: " +str(stand_laut_stapelbeschreibung))
    # print("stand_lokal (Ergebnis):        " +str(stand_lokal))
    # print("")
    
    if stand_lokal == -1:
      continue
    if val[3] > stand_lokal:
      decks_outdated[curr] = stand_lokal
      new_update_str = val[2] +": <a href='" +bk_update_link +val[1] +"'>Update " +str(val[3]) +"</a> vom " +str(val[0]) #datetime.datetime.strptime(val[0], "%Y-%m-%d").strftime("%d.%m.%Y")
      #decks_outdated_str += "<br>" +new_update_str
      if not val[5] in cat_outdated:
        cat_outdated[val[5]] = { curr : new_update_str }
      else:
        cat_outdated[val[5]][curr] = new_update_str
    else:
      #decks_uptodate_str += val[2] + ", "
      decks_uptodate_count += 1
      if not val[5] in cat_uptodate:
        cat_uptodate[val[5]] = { val[2] : val[3] }
      else:
        cat_uptodate[val[5]][val[2]] = val[3]

  i = 1
  for key, val in cat_outdated.items():
    decks_outdated_str += "<h4>" +key +"</h4>"
    prefix = ""
    for fach, update_str in val.items():
      decks_outdated_str += prefix + "{}) {}".format(i, update_str)
      prefix = "<br>"      
      i += 1
  
  i = 1
  #print(json.dumps(cat_uptodate, indent = 4))
  for key, val in cat_uptodate.items():
    decks_uptodate_str += "<br><br><b>" +key +":</b><br>"
    #decks_uptodate_str += "<br>".join()
    sortiert = sorted(val.keys())
    for fach in sortiert:
      version = val[fach]
      decks_uptodate_str += "<br>{}) {}".format(i, fach)
      if version > 0:
        decks_uptodate_str += " (Update: {})".format(version)
      i += 1
    
  if decks_outdated_str == u"" and decks_uptodate_str == u"": #Es wurden keine (ganzen) Basiskarten-Stapel gefunden. (Höchstens einzelne Karten aus dem Probierstapel.)
    if notify_of_result:
      showInfo(u"<h3>Es konnten leider keine Basiskarten-Stapel in deiner Sammlung gefunden werden.</h3>")
    return False

  if decks_outdated_str == u"": #Es wurden Stapel gefunden (wird ja ganz am Anfang geprüft) und alle gefundenen Stapel sind aktuell.
    if notify_of_result:
      showText(u"Alle {} gefundenen Basiskarten-Stapel sind aktuell.".format(decks_uptodate_count) +decks_uptodate_str +u"", title=u"Basiskarten Jura", type="")
    else:
      tooltip(u"Alle Basiskarten-Stapel sind aktuell.", period=5000)
    return False

  else: #Es wurden veraltete Stapel gefunden, für die Updates verfügbar sind.
    
    # if decks_uptodate_str: #Zusatz, falls es gleichzeitig noch aktuelle Stapel gibt.
      # decks_uptodate_str = "<h3><b>Noch aktuell:</b></h3>" +decks_uptodate_str
    
    dialog = askUserDialog(u"<style>a { color: rgb(185, 33, 33); text-decoration: none; }</style><h3>" +u"<b>Für {} Stapel sind Updates verfügbar.</b>&nbsp;🔔</h3>".format(len(decks_outdated.items())) +decks_outdated_str +"<p><i>Generelle Infos zum Update-Vorgang findest du <a href='https://basiskarten.de/updates'>hier</a>.</i></p>", [" Updates installieren ", " Aktuelle Stapel anzeigen ", "Abbrechen"], title="Basiskarten Jura")
    dialog.setIcon(QMessageBox.Icon.NoIcon)
    answer = dialog.run()
    if answer == " Updates installieren ":
      report = ""
      for sku, current_version in decks_outdated.items():
        dl_and_install_updates_for_sku(sku, current_version, decksRemote[sku][2]) #report += dl_and_install_updates_for_sku(sku, current_version) +"<br>"
      # if report != "":
        # showText(report, title="Basiskarten Jura")
    elif answer == " Aktuelle Stapel anzeigen ":
      showText(u"{} Basiskarten-Stapel sind aktuell.".format(decks_uptodate_count) +decks_uptodate_str +u"", title=u"Basiskarten Jura", type="")
  
  return True

def check_for_deck_updates_report():
    check_for_deck_updates() #Auch, wenn kein Update vorliegt, wird eine Antwort gegeben.

def bk_toggle_automatic_updates():
    global bk_automatic_updates
    bk_automatic_updates = not bk_automatic_updates
    save_config()
    
def bk_toggle_vv_anzeige():
    global vv_hinweis_anzeigen
    vv_hinweis_anzeigen = not vv_hinweis_anzeigen
    save_config()
    settings_datei_fuer_jv_speichern()

def update_individual_note_with_bk_info(nid, force_nid=None):
  # try:
  note = mw.col.getNote(int(nid))
  # except:
      # print ("#1 couldn't get nid " +str(nid))
      # return False  
  
  if note == None:
      print ("#2 couldn't get nid " +str(nid))
      return False

  save = False
  
  if force_nid == None:
    force_nid = str(nid)

  tag = u"Basiskarten"
  if  not tag.lower() in note.tags \
  and not tag         in note.tags:
      #mod_before = note.mod
      note.tags.append(tag)
      save = True
      #print "Added tag '"+tag+"' to nid "+nid + " (mod before: "+str(mod_before) +", mod after: " +str(note.mod) +")"
      #tags_not_present += 1
  try:
    if note['Quellen'] == "": 
        note['Quellen'] = "[basiskarten_id: " +force_nid +"]"
        save = True
        #basiskarten_id_not_present_in_quellen += 1
    elif note['Quellen'].lower().find("[basiskarten_id") == -1:
        note['Quellen'] += "\n<br>\n<br>\n[basiskarten_id: " +force_nid +"]"
        save = True
        #basiskarten_id_not_present_in_quellen += 1
  except:
    try:
        if note['<br>Quellen'] == "": 
            note['<br>Quellen'] = "[basiskarten_id: " +force_nid +"]"
            save = True
            #basiskarten_id_not_present_in_brquellen += 1
        elif note['<br>Quellen'].lower().find("[basiskarten_id") == -1:
            note['<br>Quellen'] += "\n<br>\n<br>\n[basiskarten_id: " +force_nid +"]"
            save = True
            #basiskarten_id_not_present_in_brquellen += 1
    except:
        pass
  if save:
      note.mod += 1
      note.flush()
      return True
  return False

def add_basiskarten_tag_to_bk_notes(show_result = False):
    
    print(u"add_basiskarten_tag_to_bk_notes")
    
    try:
        nids_str = http_get(bk_nids_url)
    except:
        showInfo(u"Keine Verbindung zum Server möglich!\n\n(Stelle: add_basiskarten_tag_to_bk_notes #1)", title="Butler-Addon")
        return False
    bk_nids = json.loads(nids_str)
    
    # if askfirst:
        # if not askUser(u"Es wurde noch nicht geprüft, ob die in deiner Sammlung enthaltenen Basiskarten bereits getagged (also als solche markiert) wurden. Das ist nötig, damit Updates und die Verbesserungsvorschlag-Anzeige auch bei Stapeln funktioniert, die du vor Dezember 2019 in deine Sammlung importiert hast.\n\nJetzt prüfen und Markierung nachholen?"):
            # return False
    
    #JuVo auf neueste Version updaten
    try:
        content = http_get(jura_vorlagen_stand_remote +"?beta=False")
    except:
        showInfo(u"Keine Verbindung zum Server möglich!\n\n(Stelle: add_basiskarten_tag_to_bk_notes #2)", title="Butler-Addon")
        return False
      
    try:
      if not content:
        if show_result:
          showInfo(u"Leere Antwort vom Server erhalten. Liegt nicht an dir, da stimmt gerade etwas mit meinem Server nicht. 🚧\n\n(Stelle: add_basiskarten_tag_to_bk_notes #3)", title="Butler-Addon")
      elif get_jura_vorlagen_version() < int(content.split(" ")[0]):
          if show_result and askUser(u"Um die Basiskarten in deiner Sammlung als solche zu taggen (→ Voraussetzung, um diese updaten zu können), muss die neueste Version der Jura-Vorlagen bei dir installiert sein. Jetzt updaten?"):
              update_jura_vorlagen(False)
          else:
              return False
    except:
      showInfo(u"Die vom Server empfangene Antwort konnte nicht verarbeitet werden. Liegt vermutlich nicht an dir!\n\n(Stelle: add_basiskarten_tag_to_bk_notes #4)\n\ncontent: " +content, title="Butler-Addon")
      return False
    
    i = 0
    
    #Über die nid nach Notizen suchen, die noch nicht mit Basiskarten-Tag und -ID versehen wurden.
    nids_without_bk_tag = find_notes("-[basiskarten_id")
    for nid in nids_without_bk_tag:
      if nid in bk_nids:
        if update_individual_note_with_bk_info(nid):
          i += 1

    try:
        cids_str = http_get(bk_cids_and_nids_url)
    except:
        showInfo(u"Keine Verbindung zum Server möglich!\n\n(Stelle: add_basiskarten_tag_to_bk_notes #4)", title="Butler-Addon")
        return False
    bk_cids_and_nids = json.loads(cids_str)
    
    #Über die cid nach Notizen suchen, die noch nicht mit Basiskarten-Tag und -ID versehen wurden.
    cids_without_bk_tag = find_cards("-[basiskarten_id")
    for cid in cids_without_bk_tag:
      cid_str = str(cid)
      if cid_str in bk_cids_and_nids.keys():
        card = mw.col.getCard(cid)
        if update_individual_note_with_bk_info(str(card.nid), force_nid=str(bk_cids_and_nids[cid_str])): #TO DO: Muss die ja überschreiben
          i += 1
    
    # if not (first_start): #and i == 0):
    if show_result:
      showInfo(u"Fertig! {} Basiskarten-Notizen wurden um den Tag \"Basiskarten\" und eine eigene Basiskarten-ID ergänzt.".format(i))
    print(u"Butler-Addon: {} Basiskarten-Notizen wurden um den Tag \"Basiskarten\" und eine eigene Basiskarten-ID ergaenzt.".format(i)) #Umlaute (ä in ergänzt) bewirken anscheinend einen Crash auf Mac.
    global bk_tagged
    bk_tagged = True
    global bk_tagged2
    bk_tagged2 = True
    save_config()
    mw.reset()
    return True
    
def alternative_html_spelling(s):
  """
  Betrifft die folgenden Elemente:
  
  area, base, br, col, command, embed, hr, img, input, keygen, link, meta, param, source, track, wbr
  """  
  return s.replace(" />", ">")
  
def upath(path):
    if isWin:
        prefix = u"file:///"
    else:
        prefix = u"file://"
    return prefix + path
  
def create_delta_for(content, update):
    path = os.path.join(mw.pm.profileFolder(), "changelog_{}_upd{}_{}.html".format(update['sku'], update['version'], update['date'][:10]))
    buf = open(path, "w", encoding="utf-8")
    media_folder = os.path.join(mw.pm.profileFolder(), u"collection.media")
    buf.write(u"""<html>
    
<head>

  <meta charset="utf-8">
  <base href="./collection.media/">

  <style>

    body, td, div, p, th {
      font-size: 14pt;
    }

    ins {
        text-decoration: none;
        background-color: #d4fcbc;
    } 

    del {
        text-decoration: line-through;
        background-color: #fbb6c2;
        color: #555;
    }

    .old, .new {
      display: none;
    }

    .field {
      text-align: center;
      font-weight: bold;
      background-color: lightgrey;
    }

    body {
      max-width: 700px;
      float: left;
      font-family: Arial;
    }
    
    h1 {
      text-align: center;
    }

    .erklaerung {
      border: 1px solid grey;
      padding: 5px;
    }
    
    .error{
      font-size: 11pt;
      background-color: rgb(255, 255, 127);
    }

  </style>

</head>

<body>

[CONTENT_PLACEHOLDER]

</body>

<script>

/**
 * htmldiff.js a diff algorithm that understands HTML, and produces HTML in the browser.
 * 
 * Autor:   https://github.com/tnwinc
 * Github:  https://github.com/tnwinc/htmldiff.js
 * License: htmldiff.js is licensed under the MIT License.
 *
 */

!function(){var e,n,t,r,i,f,_,a,o,s,u,h,l,c,d,b,p;o=function(e){return">"===e},s=function(e){return"<"===e},h=function(e){return/^\s+$/.test(e)},u=function(e){return/^\s*<[^>]+>\s*$/.test(e)},l=function(e){return!u(e)},e=function(){return function(e,n,t){this.start_in_before=e,this.start_in_after=n,this.length=t,this.end_in_before=this.start_in_before+this.length-1,this.end_in_after=this.start_in_after+this.length-1}}(),a=function(e){var n,t,r,i,f,_;for(f="char",t="",_=[],r=0,i=e.length;r<i;r++)switch(n=e[r],f){case"tag":o(n)?(t+=">",_.push(t),t="",f=h(n)?"whitespace":"char"):t+=n;break;case"char":s(n)?(t&&_.push(t),t="<",f="tag"):/\s/.test(n)?(t&&_.push(t),t=n,f="whitespace"):/[\w\#@]+/i.test(n)?t+=n:(t&&_.push(t),t=n);break;case"whitespace":s(n)?(t&&_.push(t),t="<",f="tag"):h(n)?t+=n:(t&&_.push(t),t=n,f="char");break;default:throw new Error("Unknown mode "+f)}return t&&_.push(t),_},f=function(n,t,r,i,f,_,a){var o,s,u,h,l,c,d,b,p,g,w,v,k,m,y;for(s=i,o=_,u=0,w={},c=h=m=i,y=f;m<=y?h<y:h>y;c=m<=y?++h:--h){for(k={},d=0,b=(p=r[n[c]]).length;d<b;d++)if(!((l=p[d])<_)){if(l>=a)break;null==w[l-1]&&(w[l-1]=0),v=w[l-1]+1,k[l]=v,v>u&&(s=c-v+1,o=l-v+1,u=v)}w=k}return 0!==u&&(g=new e(s,o,u)),g},d=function(e,n,t,r,i,_,a,o){var s;return null!=(s=f(e,0,t,r,i,_,a))&&(r<s.start_in_before&&_<s.start_in_after&&d(e,n,t,r,s.start_in_before,_,s.start_in_after,o),o.push(s),s.end_in_before<=i&&s.end_in_after<=a&&d(e,n,t,s.end_in_before+1,i,s.end_in_after+1,a,o)),o},r=function(e){var n,t,r,i,f,_;if(null==e.find_these)throw new Error("params must have find_these key");if(null==e.in_these)throw new Error("params must have in_these key");for(r={},n=0,i=(f=e.find_these).length;n<i;n++)for(r[_=f[n]]=[],t=e.in_these.indexOf(_);-1!==t;)r[_].push(t),t=e.in_these.indexOf(_,t+1);return r},_=function(e,n){var t,i;return i=[],t=r({find_these:e,in_these:n}),d(e,n,t,0,e.length,0,n.length,i)},n=function(n,t){var r,i,f,a,o,s,u,h,l,c,d,b,p,g,w,v;if(null==n)throw new Error("before_tokens?");if(null==t)throw new Error("after_tokens?");for(w=g=0,p=[],r={"false,false":"replace","true,false":"insert","false,true":"delete","true,true":"none"},(d=_(n,t)).push(new e(n.length,t.length,0)),a=f=0,h=d.length;f<h;a=++f)"none"!==(i=r[[w===(c=d[a]).start_in_before,g===c.start_in_after].toString()])&&p.push({action:i,start_in_before:w,end_in_before:"insert"!==i?c.start_in_before-1:void 0,start_in_after:g,end_in_after:"delete"!==i?c.start_in_after-1:void 0}),0!==c.length&&p.push({action:"equal",start_in_before:c.start_in_before,end_in_before:c.end_in_before,start_in_after:c.start_in_after,end_in_after:c.end_in_after}),w=c.end_in_before+1,g=c.end_in_after+1;for(v=[],u={action:"none"},o=function(e){return"equal"===e.action&&(e.end_in_before-e.start_in_before==0&&/^\s$/.test(n.slice(e.start_in_before,+e.end_in_before+1||9e9)))},s=0,l=p.length;s<l;s++)o(b=p[s])&&"replace"===u.action||"replace"===b.action&&"replace"===u.action?(u.end_in_before=b.end_in_before,u.end_in_after=b.end_in_after):(v.push(b),u=b);return v},t=function(e,n,t){var r,i,f,_,a,o;for(_=void 0,f=i=0,a=(n=n.slice(e,+n.length+1||9e9)).length;i<a&&(o=n[f],!0===(r=t(o))&&(_=f),!1!==r);f=++i);return null!=_?n.slice(0,+_+1||9e9):[]},p=function(e,n){var r,i,f,_,a;for(_="",f=0,r=n.length;;){if(f>=r)break;if(i=t(f,n,l),f+=i.length,0!==i.length&&(_+="<"+e+">"+i.join("")+"</"+e+">"),f>=r)break;f+=(a=t(f,n,u)).length,_+=a.join("")}return _},(c={equal:function(e,n,t){return n.slice(e.start_in_before,+e.end_in_before+1||9e9).join("")},insert:function(e,n,t){var r;return r=t.slice(e.start_in_after,+e.end_in_after+1||9e9),p("ins",r)},delete:function(e,n,t){var r;return r=n.slice(e.start_in_before,+e.end_in_before+1||9e9),p("del",r)}}).replace=function(e,n,t){return c.delete(e,n,t)+c.insert(e,n,t)},b=function(e,n,t){var r,i,f,_;for(_="",r=0,i=t.length;r<i;r++)f=t[r],_+=c[f.action](f,e,n);return _},(i=function(e,t){var r;return e===t?e:(e=a(e),t=a(t),r=n(e,t),b(e,t,r))}).html_to_tokens=a,i.find_matching_blocks=_,_.find_match=f,_.create_index=r,i.calculate_operations=n,i.render_operations=b,"function"==typeof define?define([],function(){return i}):"undefined"!=typeof module&&null!==module?module.exports=i:"undefined"!=typeof window&&(window.htmldiff=i)}();

var olds    = document.getElementsByClassName("old")
var news    = document.getElementsByClassName("new")
var outputs = document.getElementsByClassName("output")

//console.log(olds)

for (var i = 0; i < olds.length; i++) {
	var old_html = olds[i].innerHTML
	var new_html = news[i].innerHTML
	outputs[i].innerHTML = htmldiff(old_html, new_html)
}

</script>

</html>""".replace("[CONTENT_PLACEHOLDER]", content) )
    buf.close()
    open_file(upath(path))

#Aktuellen Stand eines Stapels in der Stapel-Beschreibung aller Basiskarten-Stapel mit der entsprechenden SKU speichern
def save_deck_status(sku, new_version, new_date):

  #Find best way to loop through decks
  old_way = False
  try:
    all_decks = mw.col.decks.all_names_and_ids()
    #print('Butler-Addon: Using new way to access decks. Line: ' +str( sys._getframe().f_lineno))
  except:
    #print('Butler-Addon: Using OLD way to access decks. Line: ' +str( sys._getframe().f_lineno))
    old_way = True
    all_decks = mw.col.decks.decks.items()
  
  print("\nMaking sure all decks with sku {} are up to date (version: {}, date: {}).\n===".format(sku, new_version, new_date))
  if str(new_version) == "0":
    correct_desc = """<center><a href=http://basiskarten.de>basiskarten.de</a> - Anki-Karteikarten für dein Prädikatsexamen<br>
<a href="http://basiskarten.de/{}">Hier</a> kannst du dieses Deck bewerten.</center>""".format(sku)
  elif new_date == "":
    correct_desc = """<center>Stand: Update <span id=basiskarten_version>{}</span></center><br><center><a href=http://basiskarten.de>basiskarten.de</a> - Anki-Karteikarten für dein Prädikatsexamen<br>
<a href="http://basiskarten.de/{}">Hier</a> kannst du dieses Deck bewerten.</center>""".format(new_version, sku)
  else:
    correct_desc = """<center>Stand: Update <span id=basiskarten_version>{}</span> (<span id=basiskarten_stand>{}</span>)</center><br><center><a href=http://basiskarten.de>basiskarten.de</a> - Anki-Karteikarten für dein Prädikatsexamen<br>
<a href="http://basiskarten.de/{}">Hier</a> kannst du dieses Deck bewerten.</center>""".format(new_version, new_date, sku)
  
  for deck in all_decks: #Alle Stapel durchgehen
    
    if not old_way:
      did = str(deck.id)
      deck_json = mw.col.decks.get(deck.id)
    else:
      did = deck[0]
      deck_json = mw.col.decks.decks[did]

    bk_name = deck_id_in_desc(deck_json["desc"])
    
    if bk_name == sku: # Stapel gefunden
    
      #Wenn Version/Datum gar nicht vorhanden oder eine der beiden Angaben falsch ist, Beschreibung komplett ersetzen. ✔️

      m = re.search('<span id=basiskarten_version>(\d+)<\/span>', deck_json["desc"])
      if m == None or str(m.group(1)) != str(new_version):
        deck_json["desc"] = correct_desc
        mw.col.decks.save(deck_json)
        print ("- Deck {} aktualisiert auf Version {} ({})".format(deck_json['name'], new_version, new_date))
        continue

      m = re.search('<span id=basiskarten_stand>(\d+-\d+-\d+)+<\/span>', deck_json["desc"])
      if m == None or str(m.group(1)) != str(new_date):
        deck_json["desc"] = correct_desc
        mw.col.decks.save(deck_json)
        print ("- Deck {} aktualisiert auf Version {} ({})".format(deck_json['name'], new_version, new_date))
    
  mw.reset()
  return

def install_update(update): #Update muss im JSON-Format sein.

  #tooltip(u"Installiere Update {} für {}. Das kann bis zu 15 sec dauern.".format(update['version'], update['stapel']), period=5000)

  nids_not_contained_in_collection = []
  wrong_field_count = []
  updated = {}
  conflict = {}
  user_updated_himself = {}
  note_already_updated = []
  comparison = ""
  reset_cards = {}
  content = """<div class=old>{}</div>
<div class=new>{}</div>
<p class=field>{}</p>
<div class=output></div>"""
  
  #nids durchgehen, die im Update enthalten sind.
  for n in update['nids']:

    # Hier prüfen wir, ob die jeweilige Notiz auch in der Sammlung des Users enthalten ist.
    
    # Zunächst suchen wir nach der Basiskarten-ID (nur im Feld Quellen)
    try:
      note = mw.col.getNote(find_notes("\"Quellen:*[basiskarten_id: {}]*\"".format(n['nid']))[0])
      # Früher: find_notes("\"[basiskarten_id\: {}]\"".format(n['nid']))

      # TO DO: Prüfen, ob nur eine Notiz gefunden wurde.
      
    except:
    
      # Es wurde keine Notiz mit der richtigen Basiskarten-ID gefunden, daher suchen wir direkt nach der nid und prüfen, ob es eine Basiskarte ist.
      
      try:
        print("Notiz über Basiskarten-ID nicht gefunden. Suche direkt über nid {}.".format(n['nid']))
        note = mw.col.getNote(int(n['nid']))
        if not ("Basiskarten" in note.tags or "basiskarten" in note.tags):
          print("nid gefunden, aber Basiskarten-Tag fehlt!")
          raise
          
      except:
        # Wenn das nicht klappt, notieren wir uns, dass die Karte nicht gefunden werden konnte.
        # TO DO: Karte selbst erstellen.
        # P: In welchem Stapel?
        # P: Klappt nicht, wenn Notiztyp nicht vorhanden/verändert wurde.
      
        print("Missing nid: " +str(n['nid']))
        nids_not_contained_in_collection.append(n['nid'])
        
        comparison += "<center><h3><a href='https://basiskarten.anvil.app/#?nid={nid}' target=_blank>{nid}</a> (<a href='anki://\"Quellen:*[basiskarten_id: {nid}]*\"' target=_blank>lokal</a>)</h3></center><p><span class=error>❌ Nicht geupdatet, da Notiz nicht gefunden ❌</span></p></center>".format(nid = n['nid'])
        for i, f_neu in enumerate(n['notizinhalt_neu']):
          f_alt  = n['notizinhalt_alt'][i]
          f_name = n['feld_namen'][i]
          comparison += content.format(f_alt.replace("&nbsp;", " "), f_neu.replace("&nbsp;", " "), f_name)          
        comparison += "<hr><br><br>"
        
        continue
    
    # Prüfen, welche Versionsnummer die jeweilige Notiz hat. Falls Update bereits eingespielt, diese Notiz überspringen.
    if get_basiskarten_update_version(note) >= update['version']:
      note_already_updated.append(n['nid'])
      continue
    
    #Prüfen, ob die Notiz in der Sammlung des Users die gleiche Anzahl an Feldern hat (Proxy für selben Notiztyp)
    if len(note.fields) != len(n['notizinhalt_neu']):
      print("===")
      
      # for f in note.fields:
        # print(f)
      print("nid:{}".format(n['nid']))
      print("Erwartet: {} Felder".format(len(n['notizinhalt_neu']))) #Erwartet: {} ({} Felder) => Notiztyp angeben
      print("Erhalten: {} Felder".format(len(note.fields)))
        
      print("===\n\n")
      wrong_field_count.append(n['nid'])
      
      comparison += "<center><h3><a href='https://basiskarten.anvil.app/#?nid={nid}' target=_blank>{nid}</a> (<a href='anki://\"Quellen:*[basiskarten_id: {nid}]*\"' target=_blank>lokal</a>)</h3></center><p><span class=error>❌ Nicht geupdatet, da falsche Feldanzahl (anderer Notiztyp?) ❌</span></p></center>".format(nid = n['nid'])
      for i, f_neu in enumerate(n['notizinhalt_neu']):
        f_alt  = n['notizinhalt_alt'][i]
        f_name = n['feld_namen'][i]
        comparison += content.format(f_alt.replace("&nbsp;", " "), f_neu.replace("&nbsp;", " "), f_name)          
      comparison += "<hr><br><br>"
      
      continue

    note_before_update = list(note.fields)
    
    #Zwischenstand: Die Notiz ist vorhanden, wurde noch nicht geupdatet und hat die richtige Anzahl an Feldern.
    
    #Felder der neuen Notiz-Version durchgehen
    for i, f_neu in enumerate(n['notizinhalt_neu']):
      f_alt = n['notizinhalt_alt'][i]
      
      #Felder überspringen, die sich im Update (im Vergleich zur letzten Version) NICHT geändert haben
      if f_neu == f_alt:
        continue
      #else:
      #  showInfo("Update Feld:" +f_neu +"#" +f_alt)
      
      #Eigentlicher Update-Vorgang:
      
      # vergleichs_tupel_alt = (f_alt, alternative_html_spelling(f_alt))
      # vergleichs_tupel_neu = (f_neu, alternative_html_spelling(f_neu))
      f_aktuell = alternative_html_spelling(note.fields[i])
      
      #User hat dieses Feld bereits selbst auf die neue Version aktualisiert.
      if f_aktuell == alternative_html_spelling(f_neu): #note.fields[i] in vergleichs_tupel_neu or alternative_html_spelling(note.fields[i]) in vergleichs_tupel_neu:
        # print(u"*** NID: {} ***".format(n['nid']))
        # print(u"Feld aktuell:")
        # print(note.fields[i])
        # print(u"")
        # print(u"Feld frueher (f_alt):")
        # print(f_alt)
        # print("")
        # print(u"Feld neu (f_neu):")
        # print(f_neu)
        # print("")
        # print("")
        # print("")
        try:
          user_updated_himself[str(n)] += 1
        except:
          user_updated_himself[str(n)] = 1
          
      #User hat das Feld nicht angepasst. => Einfach von alt auf neu aktualisieren.
      elif f_aktuell == alternative_html_spelling(f_alt): #note.fields[i] in vergleichs_tupel_alt or alternative_html_spelling(note.fields[i]) in vergleichs_tupel_alt:
        # print(u"Feld aktuell:")
        # print(note.fields[i])
        # print(u"")
        # print(u"Feld frueher (f_alt):")
        # print(f_alt)
        # print("")
        # print(u"Feld neu (f_neu):")
        # print(f_neu)
        # print("")
        # print("")
        # print("")
      
        note.fields[i] = f_neu
        note.flush()
        mw.col.db.execute("update notes set mod = ? where id = ?", n['mod'], n['nid'])
        try:
          updated[str(n)] += 1
        except:
          updated[str(n)] = 1
      
      #User hat das selbst anders verändert. => Änderungskonflikt
      else:
        print(u"Änderungskonflikt bei nid {}".format(n['nid']))
        print(u"\nFeld aktuell:")
        print(note.fields[i])
        print(u"\nFeld frueher (f_alt):")
        print(f_alt)
        print(u"\nFeld neu (f_neu):")
        print(f_neu)
        print("\n\n\n")
        
        if f_neu == "":
          f_neu = "[Das Feld ist in der aktuellen Version leer.]"
      
        note.fields[i] = f_neu +u"<br><br>=== Früherer Feldinhalt (vor Installation des Updates {} vom {}): ===<br><br>".format(update['version'], update['date'][:10]) + note.fields[i]
        note.flush()
        mw.col.db.execute("update notes set mod = ? where id = ?", n['mod'], n['nid'])
        try:
          conflict[str(n)] += 1
        except:
          conflict[str(n)] = 1
    
    if "reset_cards" in n:
      reset_cards[str(n['nid'])] = n['reset_cards']
    
    field_names = []
    for fn in note._fmap:
      field_names.append(fn)
      
    comparison += "<center><h3><a href='https://basiskarten.anvil.app/#?nid={nid}' target=_blank>{nid}</a> (<a href='anki://\"Quellen:*[basiskarten_id: {nid}]*\"' target=_blank>lokal</a>)</h3></center>".format(nid = n['nid'])
    for i, current in enumerate(note.fields):
      comparison += content.format(note_before_update[i].replace("&nbsp;", " "), current.replace("&nbsp;", " "), field_names[i])
    comparison += "<hr><br><br>"

  #Speichern, dass das Update eingespielt wurde - Es sei denn, keine einzige Veränderung konnte eingespielt werden.
  changed_notes = updated.copy()
  changed_notes.update(user_updated_himself)
  changed_notes.update(conflict)
  successful_notes_count = len ( changed_notes.keys() ) + len(note_already_updated)
  failed_notes_count = len(wrong_field_count) + len(nids_not_contained_in_collection)
  
  for nid, card_types in reset_cards.items():
    for ct in card_types:
      forget_cards(find_cards("\"Quellen:*[basiskarten_id: {}]*\" \"card:{}\"".format(nid, ct)))
  
  if successful_notes_count > 0:
    save_deck_status( update['sku'], update['version'], update['date'][:10] )
  mw.reset()
  
  zsf = """<font color=green>{}</font> Notizen waren bereits aktuell ✅<br>
<font color=green>{}</font> Felder wurden aktualisiert von alt auf neu ✅<br>
<font color=green>{}</font> Felder hattest du bereits selbst aktualisiert ✅<br>
<font color=blue>{}</font> Felder wurden ergänzt um die neue Version, da ihr Inhalt vor dem Update angepasst wurde 🔄<br>
<font color=orange>{}</font> Notizen nicht geupdatet, da Notiztyp geändert ❌<br>
<font color=red>{}</font> Notizen nicht in deiner Sammlung gefunden ❌""".format(len(note_already_updated), sum(updated.values()), sum(user_updated_himself.values()), sum(conflict.values()), len(wrong_field_count), len(nids_not_contained_in_collection))
  
  link_to_update_url = "https://basiskarten.anvil.app/#?s={}&v={}".format(update['sku'], update['version'])
  link_to_update_text = "<p align=center>Das aktualisierte PDF-Skript des Stapels kannst du <a href='{}' target='_blank'>hier herunterladen</a>.</p>".format(link_to_update_url)
  
  showInfo(u"<div style=\"font-size: 16px\"><b>Ergebnis für {} – Update {} vom {}:</b><br>{} von {} Notizen wurden erfolgreich geupdatet. Im Einzelnen:<br><br>".format(update['stapel'], update['version'], update['date'][:10], successful_notes_count, len(update['nids'])) +zsf)
  
  if len(nids_not_contained_in_collection) > 0:
    zsf += u"</p><p>👉 <b>Tipp:</b> Notizen, die nicht in deiner Sammlung gefunden wurden (weil sie gelöscht wurden oder erst in diesem Update neu hinzugekommen sind), kannst du erhalten, indem du zusätzlich die aktuelle Version des Stapels als Anki-Datei (APKG) <a href='{}' target='_blank'>herunterlädst</a> und in deine Sammlung importierst. Dadurch werden nur die fehlenden Notizen wiederhergestellt, sonst ändert sich nichts an deiner Sammlung.".format(link_to_update_url)
  
  if len(reset_cards) > 0:
    zsf += u"</p><p>🆕 Karten, deren Inhalt sich in dem Update grundlegend geändert hat, wurden automatisch für dich zurückgesetzt und erscheinen jetzt als neue Karten."

  erklaerung = u"""<p class=erklaerung>In diesem Dokument sind alle Notizen aufgeführt, die durch das Update verändert wurden oder hätten werden sollen. Text, der neu hinzugefügt wurde, ist <ins>grün hinterlegt</ins>. Was gelöscht wurde, ist <del>rot hinterlegt</del> und durchgestrichen.<br><br>Wenn eine Notiz aufgeführt ist, in der keine Änderungen erkennbar sind, kann das daran liegen, dass nur (unsichtbarer) HTML-Code geändert wurde, oder daran, dass du die jeweilige Änderung bereits zuvor selbst vorgenommen hast (wenn z.B. nur ein einzelner Buchstabe oder eine Norm falsch war).<br><br>Um diesen oder andere Changelogs wiederzufinden, klick in Anki auf: Butler-Addon > Sonstiges > Changelog-Ordner öffnen</p>
<h2>Ergebnis</h2>
{} von {} Notizen wurden erfolgreich geupdatet.
<p class=erklaerung>{}</p>
<h2>Veränderungen</h2>  """.format(successful_notes_count, len(update['nids']), zsf)

  comparison = "<h1>{} - Update {} ({})</h1>".format(update["stapel"], update['version'], update['date'][:10]) +link_to_update_text +erklaerung +comparison
  create_delta_for(comparison, update)
  
  if successful_notes_count == 0:
    return #"{} - Update {} ({}): ❌ Update nicht installiert ({} Notizen nicht gefunden oder Notiztyp geändert)".format(update["stapel"], update['version'], update['date'][:10], len(wrong_field_count) + len(nids_not_contained_in_collection))
  
  if failed_notes_count > 0:
    return #"{} - Update {} ({}): ⚠ Update teilweise installiert ({} Notizen nicht gefunden oder Notiztyp geändert)".format(update["stapel"], update['version'], update['date'][:10], len(wrong_field_count) + len(nids_not_contained_in_collection))
  
  return #"{} - Update {} ({}): ✔️ Update vollständig installiert.".format(update["stapel"], update['version'], update['date'][:10])
  
def read_file(filepath):
  with open(filepath, 'r', encoding="utf-8-sig") as file:
    return file.read()
    
def save_file(filepath, str):
  file = open(filepath, 'w', encoding="utf-8")
  file.write(str)
  file.close()

def replace_single_line_in_str(str, condition, replacement):
  lines     = str.split("\n")
  lines_new = lines
  i = 0
  for l in lines:
      if l.find(condition) > -1:
          #print("Found!")
          lines_new[i] = replacement
          #print("...and REPLACED!")
      i += 1
  return u"\n".join(lines_new)

def datei_loeschen(datei):
  print("Trying to delete '{}'".format(datei))
  from os.path import exists
  if not exists(datei):
    print("File does not exist.\n")
    return
  
  try:
    print("Try 1: remove")
    os.remove(datei)
    print("Success.\n")
  except:
    try:
      print("Try 2: unlink")
      os.unlink(datei)
      print("Success.\n")
    except:
      try:
        print("Try 3: pathlib")
        import pathlib
        pathlib.Path.unlink(datei)
        print("Success.\n")
      except:
        print("Unable to delete file.\n")

def settings_datei_fuer_jv_speichern():
  try:
    content = read_file("_jv.js")
  except:
    content = ""
  settings_datei_inhalt = u"""var token = '{}';
var vv_hinweis_anzeigen = {};
document.getElementById('vv_hinweis_anzeigen').innerHTML = vv_hinweis_anzeigen;
document.getElementById('token').innerHTML = token;""".format(token, str(vv_hinweis_anzeigen).lower())
  if content != settings_datei_inhalt:
    save_file("_jv.js", settings_datei_inhalt)

def set_token():
  global token
  
  showInfo(u"""<html><style>a { color: rgb(185, 33, 33); }</style><div>Um Updates herunterzuladen, benötigst du ein <b>Token</b>.<br><br>

<b><a href="https://basiskarten.anvil.app/_/api/get_token/">Klick hier</a></b>, um dein persönliches Token zu erhalten.</center><br><br>

Wenn du dein Token hast, klick auf OK. Im nächsten Schritt kannst du es dann eingeben.</div></html>""", title="Basiskarten")
  antwort = getText(u"""Dein Token:""", default=token)
  if antwort[1] == 0:
    return False
  else:
    token = antwort[0].strip()
    save_config()
    settings_datei_fuer_jv_speichern()

def dl_and_install_updates_for_sku(sku, current_version, full_name):
  
  if not bk_tagged or not bk_tagged2:
    if not add_basiskarten_tag_to_bk_notes(show_result=True):
      showInfo(u"Update konnte nicht durchgeführt werden, da die Basiskarten nicht markiert werden konnten. Bitte bei mir (thomas@basiskarten.de) melden, damit ich mir anschauen kann, was das Problem ist. :-)")
      return #sku +": Update nicht installiert, da die Basiskarten nicht getagged werden konnten." +"<br>"
  
  #Falls Token leer, nachfragen!
  if not token:
    set_token()
    if not token:
      return #sku +": Update nicht installiert, da kein Token vorhanden." +"<br>"
      
  #tooltip(u"Installiere Updates für {}. Bitte warten...".format(full_name), period=5000)
  
  params = {
    'sku' : sku,
    'current_version' : current_version,
    'token' : token
  }
  try:
    updates = http_post(bk_dl_updates_link, params)
  except:
    showInfo(u"Server nicht erreichbar! Besteht eine Internetverbindung?", title=u"Butler-Addon")
    return #sku +": Update nicht installiert, da Server nicht erreichbar." +"<br>"
  
  try:
    updates = json.loads(updates)
  except:
    showInfo(u"<html>Kein Update erhalten! Grund:<p>{}</p></html>".format(updates), title=u"Update für " +full_name)
    return #sku +": Update nicht installiert. Grund: " +str(updates) +"<br>"
  
  result = ""
  for u in updates:
    install_update(u) #result += sku +": " +install_update(u) +"<br>"
  
  return #result




######################################
#
#   NACHRICHTEN
#
######################################  

def check_for_message(info_all = None, notify_of_result = True):
    
    print("Checking if new messages are available (Butler-Addon).")
    
    if info_all == None:
      info_all = get_info(notify_of_result = notify_of_result)
    if info_all == None:
      return False
    
    stand_remote = info_all['news_date']
    global nachrichten_stand_lokal
    
    if nachrichten_stand_lokal < stand_remote or notify_of_result:
      showInfo(info_all['news_msg'], title="Butler-Addon - Nachricht vom " +stand_remote +" Uhr")
      nachrichten_stand_lokal = stand_remote
      save_config()

def check_for_message_report():
    check_for_message()

def nachrichten_toggle_automatic_updates():
    global nachrichten_empfangen
    nachrichten_empfangen = not nachrichten_empfangen
    save_config()



######################################
#
#   EMPFOHLENE ADD-ONS
#
######################################

#Welchen Stand haben die Addons lokal?
def get_recommended_plugins_stand_lokal():
    pfad = os.path.dirname(os.path.realpath(__file__))
    plugin_ordner = pfad[:pfad.rindex(pfad_trenner)]
    try:
        with open(plugin_ordner + pfad_trenner + "druck_plugin_fuer_die_jura_vorlagen" + pfad_trenner + 'version_empfohlene_plugins.txt', 'r') as file:
            return file.read()
    except:
        return ""

#Welchen Stand sollten die Addons laut Server haben?
def download_recommended_plugins(silent = False, beta = False):
    global recommended_plugins_stand_lokal
    
    pfad = os.path.dirname(os.path.realpath(__file__))
    if isWin:
        plugin_ordner = pfad[:pfad.rindex("\\")]
    else:
        plugin_ordner = pfad[:pfad.rindex("/")]
        
    try:
        content = http_get(recommended_plugins_stand_remote_21 +"?beta=" +str(beta))
    except:
        showInfo(u"Keine Verbindung zum Server möglich!\n\n(Stelle: download_recommended_plugins)", title="Butler-Addon")
        return False
    if beta and content.find("Beta") == -1:
        showInfo(u"Aktuell ist keine Beta-Version verfügbar.", title="Butler-Addon")
        return
    
    if dl_file_and_unzip_to(recommended_plugins_21 +"?beta=" +str(beta) +"&email=" +beta_email, plugin_ordner):

        recommended_plugins_stand_lokal = content[:16]
        save_config()

        if not silent:
            if beta:
              showInfo(u"Beta-Version der empfohlenen Addons installiert. Bitte Anki einmal neustarten!\n\n(👉 Tipp: Falls etwas nicht klappen sollte, kannst du jederzeit wieder die stabile Version installieren. Klick dazu einfach auf: Butler-Addon > Empfohlene Addons > Letzte stabile Version laden)", title="Butler-Addon")
            else:
              showInfo(u"Empfohlene Addons installiert. Bitte Anki einmal neustarten!", title="Butler-Addon")

def download_recommended_plugins_ask_first():
    if askUser(u"Letzte stabile Version der <font color=green><b>empfohlenen Addons</b></font> installieren? Welche das sind und was sie tun, <a href=\"" +recommended_plugins_info +u"\">kannst du hier nachlesen</a>.", title="Butler-Addon"):
        download_recommended_plugins()
        
def download_recommended_plugins_beta_ask_first():
    if askUser(u"Letzte Beta-Version der <font color=green><b>Empfohlene Addons</b></font> installieren? Welche das sind und was sie tun, <a href=\"" +recommended_plugins_info +u"\">kannst du hier nachlesen</a>.", title="Butler-Addon"):
        download_recommended_plugins(beta = True)
    
def check_for_recommended_plugins_update(info_all = None, notify_of_result = True):
    
    print("Checking if recommendend plugins are up to date (Butler-Addon).")
    
    install = False
    
    if info_all == None:
      info_all = get_info(notify_of_result = notify_of_result)
    if info_all == None:
      return False
    
    #Prüfen, ob empfohlene Addons überhaupt installiert
    if get_recommended_plugins_stand_lokal() == "":
      if notify_of_result:
        if askUser(u"Bisher ist noch gar keine Version der <font color=green><b>empfohlenen Addons</b></font> installiert. (Welche das sind und was sie tun, <a href=\"" +recommended_plugins_info +u"\">kannst du hier nachlesen</a>.)<br><br>Jetzt installieren?", title="Butler-Addon"):
          install = True
      return False  

    #Lokalen Stand ermitteln
    try:
        stand_lokal_date = datetime.datetime.strptime(recommended_plugins_stand_lokal, "%Y-%m-%d %H:%M")
    except:
        stand_lokal_date = datetime.datetime.strptime(recommended_plugins_stand_lokal, "%Y-%m-%d")
    
    #Remote Stand ermitteln
    if beta_participate:
      remote_stand_str   = info_all['rec_addons']['beta_v']
      remote_stand_date  = datetime.datetime.strptime(remote_stand_str, "%Y-%m-%d %H:%M")
      release_notes      = info_all['rec_addons']['beta_rel_notes']
    else:
      remote_stand_str   = info_all['rec_addons']['stable_v']
      remote_stand_date  = datetime.datetime.strptime(remote_stand_str, "%Y-%m-%d %H:%M")
      release_notes      = info_all['rec_addons']['stable_rel_notes']
    
    #Vergleichen
    if stand_lokal_date < remote_stand_date:   
      if askUser(u"Es gibt eine neue Version der <font color=green><b>empfohlenen Addons</b></font>! ({}) <br><br>Release Notes:<p><pre>{}</pre></p>Mehr Infos dazu <a href='{}'>hier</a>.<br><br><center>Jetzt herunterladen?</center>".format(remote_stand_str, release_notes, recommended_plugins_changelog), title="Butler-Addon"):
          download_recommended_plugins(beta = beta_participate)
    elif notify_of_result:
        showInfo(u"Deine Version der <font color=green><b>empfohlenen Addons</b></font> ist aktuell!", title="Butler-Addon")

def check_for_recommended_plugins_update_report():
    check_for_recommended_plugins_update()

def recommended_plugins_toggle_automatic_updates():
    global recommended_plugins_automatic_updates
    recommended_plugins_automatic_updates = not recommended_plugins_automatic_updates
    save_config()



######################################
#
#   JURA-VORLAGEN
#
######################################

def get_jura_vorlagen_version():
    result = -1
    
    old_way = False
    try:
      all_models = mw.col.models.all_names_and_ids()
      #print('Butler-Addon: Using new way to access models. Line: ' +str( sys._getframe().f_lineno))
    except:
      #print('Butler-Addon: Using OLD way to access models. Line: ' +str( sys._getframe().f_lineno))
      old_way = True
      all_models = mw.col.models.models.keys()
    
    for m in all_models:
    
        if not old_way:
          mid   = m.id #Integer!
          model = mw.col.models.get(mid)
        else:
          mid   = mw.col.models.models[m]['id'] #Integer!
          model = mw.col.models.models[m]
        
        if mid in {1432993954137: u"Fragenfolge", 1432994415217: u"Lückentext+", 1398877790071: u"Rechtsfrage", 1358755191829: u"Vergleichende Frage", 1358755180344: "Prüfungsschema", 1399720652554: "Streitstand"}:
            result = 8
            css = model['css']
            if css.find("Jura-Vorlagen-Version") > -1:
                lines = css.split("\n")
                for l in lines:
                    if l.find("Jura-Vorlagen-Version") > -1:
                        version_line = l.strip("/*")
                        version_line = version_line.strip("*/")
                        version_line = version_line.strip(" ")
                        return json.loads(version_line)['Jura-Vorlagen-Version']
    return result

def delete_tmp_jura_vorlagen():
  datei_loeschen("jura_vorlagen.apkg")

def download_jura_vorlagen():
    tooltip("Importiere die Jura-Vorlagen. Das kann bis zu 20 sec dauern. Bitte warten...", period=15000)
    try:
        if dl_file_to(jura_vorlagen_dl, "jura_vorlagen.apkg"):
            open_file("jura_vorlagen.apkg")
    except:
        showInfo(u"Keine Verbindung zum Server möglich! (Stelle: download_jura_vorlagen)")
        return False
    mw.progress.timer(120000, delete_tmp_jura_vorlagen, False)
    
def download_jura_vorlagen_ask_first():
    if askUser(u"Jura-Vorlagen installieren?<p><font color=red>Beachte:</font> Wenn bereits eine ältere Version der Jura-Vorlagen installiert ist, solltest du diese zunächst aktualisieren. Falls du das noch nicht getan hast, klick jetzt auf \"Nein\" und dann auf: Butler-Addon > Jura-Vorlagen > Auf die neueste Version updaten/zurücksetzen</p>"):
        download_jura_vorlagen()
        
def update_jura_vorlagen(beta):
   
    if beta:
        """diag = askUserDialog(u"Vor der Installation einer Beta-Version ist es sinnvoll, ein Backup deiner Sammlung zu machen (nur für den Fall, dass etwas schiefgeht). Das geht im Anki-Hauptmenü über Datei > Exportieren...<br><br>Zunächst ein Backup erstellen?", [_(u"Ja"), _(u"Nein, Installation fortsetzen"), _(u"Abbrechen")], title=u"Butler-Addon")
        diag.setDefault(0)
        choice = diag.run()
        if choice == u"Ja":
            showInfo(u"Alles klar. Dann geh jetzt bitte im Anki-Hauptmenü auf Datei > Exportieren...<br><br>Anschließend installierst du die neue Version der Jura-Vorlagen über Butler-Addon > Jura-Vorlagen > Jetzt prüfen, ob es eine neue Version gibt")
            return
            #aqt.exporting.ExportDialog(mw) #Macht Probleme unter Anki 2.1.29+, weil der Exporter im Hintergrund läuft und dabei die Sammlung geschlossen wird.
        elif choice == u"Abbrechen":
            return"""
            
        try:
            content = http_get(jura_vorlagen_content +"?beta=True" +"&email=" +beta_email);
        except:
            showInfo(u"Keine Verbindung zum Server möglich!\n\n(Stelle: update_jura_vorlagen #1)", title="Butler-Addon")
            return False
    else:
        try:
            content = http_get(jura_vorlagen_content +"?beta=False");
        except:
            showInfo(u"Keine Verbindung zum Server möglich!\n\n(Stelle: update_jura_vorlagen #2)", title="Butler-Addon")
            return False
    
    if not content:
        showInfo(u"Leere Antwort vom Server erhalten. Das liegt nicht an dir, da stimmt gerade etwas mit meinem Server nicht. 🚧\n\n(Stelle: update_jura_vorlagen #3)", title="Butler-Addon")
        return False
    elif content == "Error: No beta available":
        showInfo(u"Aktuell ist keine Beta-Version verfügbar.", title="Butler-Addon")
        return
    try:
      jura_vorlagen_dl = json.loads(content)
    except:
      showInfo(u"Fehler: Antwort vom Server konnte nicht verarbeitet werden. 😵\n\n(Stelle: update_jura_vorlagen #4)", title="Butler-Addon")
      return

    # 2. Installierte Notiztypen durchgehen. Entspricht eine der MIDs den Jura-Vorlagen? Wenn ja, dann durch die neuen Angaben ersetzen.
    result = ""
    
    #Optimalen Weg finden, um Notiztypen zu durchlaufen
    old_way = False
    try:
      all_models = mw.col.models.all_names_and_ids()
      print('Butler-Addon: Using new way to access models. Line: ' +str( sys._getframe().f_lineno))
    except:
      print('Butler-Addon: Using OLD way to access models. Line: ' +str( sys._getframe().f_lineno))
      old_way = True
      all_models = mw.col.models.models.keys()
    
    for m in all_models:
    
        if not old_way:
          mid   = str(m.id) #Achtung: mid ist hier ein String!
          model = mw.col.models.get(mid)
        else:
          mid   = m #Achtung: mid ist hier ein String!
          model = mw.col.models.models[m]

        if mid in jura_vorlagen_dl:
            result += model['name'] +u"  ➡  "
            model['css']   = jura_vorlagen_dl[mid]['css']
            model['flds']  = jura_vorlagen_dl[mid]['flds']
            
            #mw.col.models.models[m]['mod']   = jura_vorlagen_dl[m]['mod'] #geht nicht mehr seit Anki 2.1.28. Stattdessen:
            model['mod']   += 1 #Notwendig, weil seit Anki 2.1.28 sonst dieser "attempt to save stale template"-Fehler kommt.
            model['name']  = jura_vorlagen_dl[mid]['name']
            model['tmpls'] = jura_vorlagen_dl[mid]['tmpls']
            
            mw.col.models.save(model)

            result += model['name'] +"\n"
# Disabled since Anki 2.1.29, see: https://forums.ankiweb.net/t/anki-2-1-29-win-mac-attributeerror-nonetype-object-has-no-attribute-cancel/1598    mw.progress.finish()

    # 3. Ergebnis präsentieren
    if not result:
        showInfo(u"Keine Jura-Vorlagen gefunden, die man updaten könnte!")
    else:
        mw.reset()
        
        # Zusatzdateien herunterladen (z.B. um Normen zu verlinken)
        if beta:
            dl_file_and_unzip_to(jura_vorlagen_zusatz +"?beta=True", '')
        else:
            dl_file_and_unzip_to(jura_vorlagen_zusatz +"?beta=False", '')
            
        showInfo(u"Folgende Notiztypen wurden geupdatet/zurückgesetzt:\n\n"+result, title="Butler-Addon")
    
def update_jura_vorlagen_ask_first():
    if askUser(u"Jura-Vorlagen auf die letzte stabile Version updaten/zurücksetzen?"):
        update_jura_vorlagen(False)
        
def update_jura_vorlagen_beta_ask_first():
    if askUser(u"Jura-Vorlagen auf die aktuelle Beta-Version updaten?"):
        update_jura_vorlagen(True)
    
def check_for_jura_vorlagen_update(info_all = None, notify_of_result = True):

    #Prüfen, ob überhaupt installiert
    if get_jura_vorlagen_version() == -1:
        if notify_of_result and askUser(u"Die <b>Jura-Vorlagen</b> sind bisher noch gar nicht installiert! Jetzt herunterladen und installieren?"):
            download_jura_vorlagen()
        return
    
    print("Checking if Jura-Vorlagen are up to date (Butler-Addon).")
    
    #Info holen, was aktuell ist
    if info_all == None:
      info_all = get_info(notify_of_result = notify_of_result)
    if info_all == None:
      return False
    
    #Relevanten Remote-Stand ermitteln
    if beta_participate:
      remote_stand   = info_all['jv']['beta_v']
      release_notes  = info_all['jv']['beta_rel_notes']
    else:
      remote_stand   = info_all['jv']['stable_v']
      release_notes  = info_all['jv']['stable_rel_notes']

    if get_jura_vorlagen_version() < remote_stand:
        if askUser(u"Es gibt eine neue Version der <font color=red><b>Jura-Vorlagen</b></font>! ({}) <br><br>Release Notes:<p><pre>{}</pre></p>Mehr Infos dazu <a href='{}'>hier</a>.<br><br><center>Jetzt herunterladen?</center>".format(remote_stand, release_notes, jura_vorlagen_changelog), title="Butler-Addon"):
            update_jura_vorlagen(content.find("Beta") > -1)

    elif notify_of_result:
        showInfo(u"Deine Version der <b>Jura-Vorlagen</b> ist aktuell!", title="Butler-Addon")

def check_for_jura_vorlagen_update_report():
    check_for_jura_vorlagen_update()
        
def get_current_font_size(mobile = False):

    #Optimalen Weg finden, um Notiztypen zu durchlaufen
    old_way = False
    try:
      all_models = mw.col.models.all_names_and_ids()
      print('Butler-Addon: Using new way to access models. Line: ' +str( sys._getframe().f_lineno))
    except:
      print('Butler-Addon: Using OLD way to access models. Line: ' +str( sys._getframe().f_lineno))
      old_way = True
      all_models = mw.col.models.models.keys()
    
    for m in all_models:
    
        if not old_way:
          mid   = m.id #Achtung: mid ist hier Integer!
          model = mw.col.models.get(mid)
        else:
          mid   = mw.col.models.models[m]['id'] #Achtung: mid ist hier Integer!
          model = mw.col.models.models[m]

        if mid in {1432993954137: u"Fragenfolge", 1432994415217: u"Lückentext+", 1398877790071: u"Rechtsfrage", 1358755191829: u"Vergleichende Frage", 1399720652554: "Streitstand", 1358755180344: "Prüfungsschema"}:
            # print("GOING IN!")
            css = model['css'].split("\n")
            for l in css:
                if not mobile:
                    if l.find(u"/* Allgemeine Schriftgröße */") > -1 and l.find(u"font-size") > -1:
                        return str(l.split(":")[1].split("px")[0]).strip()
                else:
                    if l.find(u"/* Allgemeine Schriftgröße -- auf dem Smartphone */") > -1 and l.find(u"font-size") > -1:
                        return l.split(":")[1].split("px")[0].strip()
                        
def replace_single_line_in_css(condition, replacement):
    #Optimalen Weg finden, um Notiztypen zu durchlaufen
    old_way = False
    try:
      all_models = mw.col.models.all_names_and_ids()
      print('Butler-Addon: Using new way to access models. Line: ' +str( sys._getframe().f_lineno))
    except:
      print('Butler-Addon: Using OLD way to access models. Line: ' +str( sys._getframe().f_lineno))
      old_way = True
      all_models = mw.col.models.models.keys()
    
    for m in all_models:
    
        if not old_way:
          mid = m.id #Achtung: mid ist hier Integer!
          model = mw.col.models.get(mid)
        else:
          mid   = mw.col.models.models[m]['id'] #Achtung: mid ist hier Integer!
          model = mw.col.models.models[m]

        if mid in {1432993954137: u"Fragenfolge", 1432994415217: u"Lückentext+", 1398877790071: u"Rechtsfrage", 1358755191829: u"Vergleichende Frage", 1399720652554: "Streitstand", 1358755180344: "Prüfungsschema"}:
            # print("GOING IN!")
            css_lines     = model['css'].split("\n")
            css_lines_new = css_lines
            i = 0
            for l in css_lines:
                if l.find(condition) > -1:
                    #print("Found!")
                    css_lines_new[i] = replacement
                    #print("...and REPLACED!")
                i += 1
            css = "\n".join(css_lines_new)
            if css != model['css']:
                model['css'] = css
                mw.col.models.save(model)
    mw.reset()
        
def change_font_size_of_jura_vorlagen():
    #showInfo(get_current_font_size(True))
    antwort = getText(u"Bitte neue Standard-Schriftgröße in Pixel eingeben!", default=get_current_font_size())
    if antwort[1] == 0:
        return
    try:
        size = int(antwort[0])
        replace_single_line_in_css(u"/* Allgemeine Schriftgröße */", u" font-size: "+str(size) +u"px; /* Allgemeine Schriftgröße */")
        
        replace_single_line_in_css(u"/* Schriftgröße für Elemente wie \"Tags\", \"Aufdecken\", \"Verbesserungsvorschlag\" */", u" font-size: "+ str(size -2) +u"px; /* Schriftgröße für Elemente wie \"Tags\", \"Aufdecken\", \"Verbesserungsvorschlag\" */")
        
        replace_single_line_in_css(u"/* Schriftgröße für Tabellen, die in den Feldern einer Notiz enthalten sind (also solche Tabellen, die ein User der Jura-Vorlagen erstellt hat und die NICHT automatisch als Teil einer Jura-Vorlage generiert werden wie z.B. bei \"Streitstand\" oder \"vergleichende Frage\") */", u" font-size: "+ str(size -1) +u"px; /* Schriftgröße für Tabellen, die in den Feldern einer Notiz enthalten sind (also solche Tabellen, die ein User der Jura-Vorlagen erstellt hat und die NICHT automatisch als Teil einer Jura-Vorlage generiert werden wie z.B. bei \"Streitstand\" oder \"vergleichende Frage\") */")
        
        replace_single_line_in_css(u"/* Schriftgröße für die Streitstandtabelle */", u" font-size: "+str(size -1) +u"px; /* Schriftgröße für die Streitstandtabelle */")
        
        replace_single_line_in_css(u"/* Schriftgröße für die Vergleichstabelle */", u" font-size: "+str(size -1) +u"px; /* Schriftgröße für die Vergleichstabelle */")
        
        replace_single_line_in_css(u"/* Schriftgröße für den Beispielfall */", u" font-size: "+str(size -1) +u"px; /* Schriftgröße für den Beispielfall */")
        
        mw.reset()
        showInfo(u"Standard-Schriftgröße erfolgreich geändert!")
        
        antwort = getText(u"Bitte jetzt noch die neue Standard-Schriftgröße für die Anzeige <i>auf deinem Smartphone</i> eingeben! (Ebenfalls in Pixel)", default=get_current_font_size(mobile = True))

        if antwort[1] == 0:
            return

        size = int(antwort[0])
        
        replace_single_line_in_css(u"/* Allgemeine Schriftgröße -- auf dem Smartphone */", u" font-size: "+str(size) +u"px; /* Allgemeine Schriftgröße -- auf dem Smartphone */")
        
        replace_single_line_in_css(u"/* Schriftgröße für Elemente wie \"Tags\", \"Aufdecken\", \"Verbesserungsvorschlag\" -- auf dem Smartphone */", u" font-size: "+ str(size -2) +u"px; /* Schriftgröße für Elemente wie \"Tags\", \"Aufdecken\", \"Verbesserungsvorschlag\" -- auf dem Smartphone */")
        
        replace_single_line_in_css(u"/* Schriftgröße für Tabellen, die in den Feldern einer Notiz enthalten sind (also solche Tabellen, die ein User der Jura-Vorlagen erstellt hat und die NICHT automatisch als Teil einer Jura-Vorlage generiert werden wie z.B. bei \"Streitstand\" oder \"vergleichende Frage\") -- auf dem Smartphone */", u" font-size: "+ str(size -1) +u"px; /* Schriftgröße für Tabellen, die in den Feldern einer Notiz enthalten sind (also solche Tabellen, die ein User der Jura-Vorlagen erstellt hat und die NICHT automatisch als Teil einer Jura-Vorlage generiert werden wie z.B. bei \"Streitstand\" oder \"vergleichende Frage\") -- auf dem Smartphone */")
        
        replace_single_line_in_css(u"/* Schriftgröße für die Streitstandtabelle -- auf dem Smartphone */", u" font-size: "+str(size -1) +u"px; /* Schriftgröße für die Streitstandtabelle -- auf dem Smartphone */")
        
        replace_single_line_in_css(u"/* Schriftgröße für die Vergleichstabelle -- auf dem Smartphone */", u" font-size: "+str(size -1) +u"px; /* Schriftgröße für die Vergleichstabelle -- auf dem Smartphone */")
        
        replace_single_line_in_css(u"/* Schriftgröße für den Beispielfall -- auf dem Smartphone */", u" font-size: "+str(size -1) +u"px; /* Schriftgröße für den Beispielfall -- auf dem Smartphone */")
        
        mw.reset()
        showInfo(u"Fertig! Schriftgröße erfolgreich geändert!")
        
    except:
        showInfo(u"Das hat leider nicht geklappt: Fehler beim Ändern der Schriftgröße!")
        
def change_color_of_links():
    if get_jura_vorlagen_version() < 18:
      showInfo("Diese Funktion setzt voraus, dass die Jura-Vorlagen 18 oder höher installiert sind. Deine Version: {}".format(get_jura_vorlagen_version()))
      return

    if askUser(u"<p>Möchtest du die Linkfarbe für Anki im <b>normalen Modus</b> (schwarze Schrift auf weißem Hintergrund) oder im <b>Night-Mode</b> (weiße Schrift auf schwarzem Hintergrund, aktivierbar unter Extras > Einstellungen) ändern?<br><br>Ja = normaler Modus<br>Nein = Night-Mode</p>"):

        # Farbe ändern für alle Links:
        # Assozierter CSS-Code in Jura-Vorlagen:
        """
a {
 text-decoration: none;
 color: grey; /* Butler-Addon-Code: links_normal_color */
} 
        """
          
        antwort = getText(u"Bitte neue Linkfarbe eingeben (betrifft alle Links, nicht nur Gesetze)! Bsp: \"red\" oder \"#ff0000\"; <a href='https://www.w3schools.com/colors/colors_picker.asp'>Color Picker</a>", default="grey")
        if antwort[1] == 0:
            return
        try:
            replace_single_line_in_css(u"links_normal_color", u" color: " +antwort[0] +u"; /* Butler-Addon-Code: links_normal_color */")
            mw.reset()
            
            # Farbe ändern für Gesetzeslinks
            # Assozierter CSS-Code in Jura-Vorlagen:
            """
.gesetz {
 animation-name: fadeIn;
 animation-duration: 1s;
 color: grey; /* Butler-Addon-Code: ges_normal_color_1 */
}

@keyframes fadeIn {
 from {color: black;}
 to   {color: grey;} /* Butler-Addon-Code: ges_normal_color_2 */
}
            """        
            antwort = getText(u"Bitte jetzt die neue Farbe für automatisch verlinkte Normen eingeben (betrifft ausschließlich Gesetze)! Bsp: \"red\" oder \"#ff0000\"; <a href='https://www.w3schools.com/colors/colors_picker.asp'>Color Picker</a>", default="grey")
            if antwort[1] == 0:
                return
            replace_single_line_in_css(u"ges_normal_color_1", u" color: "       +antwort[0] +u"; /* Butler-Addon-Code: ges_normal_color_1 */")
            replace_single_line_in_css(u"ges_normal_color_2", u" to   {color: " +antwort[0] +u"} /* Butler-Addon-Code: ges_normal_color_2 */")
            mw.reset()
            showInfo(u"Fertig! Farbe erfolgreich geändert!")
        except:
            showInfo(u"Das hat leider nicht geklappt: Fehler beim Ändern der Schriftgröße!")
            
    else:
      
        # Farbe ändern für alle Links im Nightmode:
        # Assozierter CSS-Code in Jura-Vorlagen:
        """
.nightMode a {
color: #999999; /* Butler-Addon-Code: links_nightmode_color */
}
        """
          
        antwort = getText(u"Bitte neue Linkfarbe eingeben (betrifft alle Links, nicht nur Gesetze)! Bsp: \"red\" oder \"#ff0000\"; <a href='https://www.w3schools.com/colors/colors_picker.asp'>Color Picker</a>", default="#999999")
        if antwort[1] == 0:
            return
        try:
            replace_single_line_in_css(u"links_nightmode_color", u" color: " +antwort[0] +u"; /* Butler-Addon-Code: links_nightmode_color */")        
            mw.reset()
            
            # Farbe ändern für Gesetzeslinks im Nightmode
            # Assozierter CSS-Code in Jura-Vorlagen:
            """
.gesetz .nightMode {
animation-name: fadeIn_night_mode;
animation-duration: 1s;
color: grey; /* Butler-Addon-Code: ges_nightmode_color_1 */
}

@keyframes fadeIn_night_mode {
from {color: white;}
to   {color: #999999;} /* Butler-Addon-Code: ges_nightmode_color_2 */
}
            """        
            antwort = getText(u"Bitte jetzt die neue Farbe für automatisch verlinkte Normen eingeben (betrifft ausschließlich Gesetze)! Bsp: \"red\" oder \"#ff0000\"; <a href='https://www.w3schools.com/colors/colors_picker.asp'>Color Picker</a>", default="#999999")
            if antwort[1] == 0:
                return
            replace_single_line_in_css(u"ges_nightmode_color_1", u" color: "       +antwort[0] +u"; /* Butler-Addon-Code: ges_nightmode_color_1 */")
            replace_single_line_in_css(u"ges_nightmode_color_2", u" to   {color: " +antwort[0] +u"} /* Butler-Addon-Code: ges_nightmode_color_2 */")
            mw.reset()
            showInfo(u"Fertig! Farbe erfolgreich geändert!")
        except:
            showInfo(u"Das hat leider nicht geklappt: Fehler beim Ändern der Schriftgröße!")


def jura_vorlagen_toggle_automatic_updates():
    global jura_vorlagen_automatic_updates
    jura_vorlagen_automatic_updates = not jura_vorlagen_automatic_updates
    save_config()
    

def get_model(mid):
    #Optimalen Weg finden, um Notiztypen zu durchlaufen
    old_way = False
    try:
      all_models = mw.col.models.all_names_and_ids()
      print('Butler-Addon: Using new way to access models. Line: ' +str( sys._getframe().f_lineno))
    except:
      print('Butler-Addon: Using OLD way to access models. Line: ' +str( sys._getframe().f_lineno))
      old_way = True
      all_models = mw.col.models.models.keys()
    
    for m in all_models:
    
        if not old_way:
          mid   = str(m.id) #Achtung: mid ist hier ein String!
          model = mw.col.models.get(mid)
        else:
          mid   = m #Achtung: mid ist hier ein String!
          model = mw.col.models.models[m]

        if mid in jura_vorlagen_dl:
            result += model['name'] +u"  ➡  "
            model['css']   = jura_vorlagen_dl[mid]['css']
            model['flds']  = jura_vorlagen_dl[mid]['flds']
            
            #mw.col.models.models[m]['mod']   = jura_vorlagen_dl[m]['mod'] #geht nicht mehr seit Anki 2.1.28. Stattdessen:
            model['mod']   += 1 #Notwendig, weil seit Anki 2.1.28 sonst dieser "attempt to save stale template"-Fehler kommt.
            model['name']  = jura_vorlagen_dl[mid]['name']
            model['tmpls'] = jura_vorlagen_dl[mid]['tmpls']
            
            mw.col.models.save(model)

######################################
#
#   FRÜHERE TABELLEN-FORMATIERUNG ENTFERNEN
#
###################################### 
    
def remove_formatting_from_tables_in_cards():
    ids = find_notes("\"<table\" (font-size 13px) OR (width=\"28px\") OR (padding 2px)")
    
    if ids == []:
      showInfo("In deiner Sammlung gibt es keine Notizen mit alter Formatierung. 👌")
      return
    
    if not askUser(u"Alte Formatierung aus Tabellen entfernen? Betrifft potentiell {} Notizen.\n\nDadurch kannst du insbesondere die Schriftgröße dieser Tabellen (während der Abfrage) zentral über das Butler-Addon einstellen.".format(len(ids))):
        return

    changed = []
    for id in ids:
        note = mw.col.getNote(id)
        
        for i, inhalt in enumerate(note.fields):

            tmp = inhalt
            
            m = re.search('<table (border="1"\s?)+style="(.*)font-size:\s?(13px|10pt|12px);?(.*)"\s?>|<table style="(.*)font-size:\s?(13px|10pt|12px);?(.*)"\s?(border="1"\s?)+>', tmp)
            while m != None:
              style = ""
              for gn in (2, 4, 5, 7):
                if m.group(gn) != None:
                  style += " " +re.sub('(\s?width:\s?100%;?\s?|\s?border-collapse:\s?collapse;?\s?)*', '', m.group(gn))
              if style != "":
                style = ' style="' +style.strip() +'"'
              tmp = re.sub('<table (border="1"\s?)+style="(.*)font-size:\s?(13px|10pt|12px);?(.*)"\s?>|<table style="(.*)font-size:\s?(13px|10pt|12px);?(.*)"\s?(border="1"\s?)+>', "<table class=user_tabelle{}>".format(style), tmp)
              m = re.search('<table (border="1"\s?)+style="(.*)font-size:\s?(13px|10pt|12px);?(.*)"\s?>|<table style="(.*)font-size:\s?(13px|10pt|12px);?(.*)"\s?(border="1"\s?)+>', tmp)
            
            tmp = re.sub('font-size:\s?13px;?', '', tmp)
            tmp = re.sub('padding(-top|-bottom|-left|-right)?:\s?2px;?', '', tmp)
            tmp = re.sub('width="28px"', '', tmp)
            tmp = re.sub('style="\s*"',  '', tmp)

            if inhalt != tmp:
              if not str(id) in changed:
                changed.append(str(id))
              note.fields[i] = tmp
              note.flush()
    mw.reset()
    if len(changed) == 0:
      showInfo("0 Notizen angepasst. (Anpassung war nicht notwendig. 👍)")
    elif askUser("{} Notizen angepasst. Veränderte Notizen in der Kartenverwaltung anzeigen?".format(len(changed))):
      lookup_browser("nid:" +",".join(changed))



######################################
#
#   FUNKTIONEN FÜR DIESES PLUGIN
#
######################################

def reset_values():
    #Einstellungen für das Butler-Addon:
    global first_start
    first_start                    = True
    global butler_addon_automatic_updates
    butler_addon_automatic_updates = True

    #Einstellungen und Links für die empfohlenen Addons:
    global recommended_plugins_automatic_updates
    recommended_plugins_automatic_updates = True

    #Einstellungen und Links zum Updaten der Basiskarten:
    global bk_automatic_updates
    bk_automatic_updates = True
    
    global bk_tagged
    bk_tagged = False
    global bk_tagged2
    bk_tagged2 = False

    #Einstellungen und Links für die Jura-Vorlagen:
    global jura_vorlagen_automatic_updates
    jura_vorlagen_automatic_updates = True

    #Teilnahme_am_Beta-Programm
    global beta_asked
    beta_asked = False
    global beta_participate
    beta_participate = False
    global beta_email
    beta_email = u""

    #Nachrichten
    global nachrichten_stand_lokal
    nachrichten_stand_lokal = "2000-01-01 00:00"
    global nachrichten_empfangen
    nachrichten_empfangen = True
    
def dl_beta_png():
    beta_png_path = os.path.dirname(os.path.realpath(__file__)) +pfad_trenner +"beta.png"
    if not os.path.exists(beta_png_path):
        try:
            dl_file_to("http://thomaskahn.de/anki-butler-addon/updates/beta.png", beta_png_path)
        except:
            return False

def ask_on_first_start():
    if first_start:
        if askUser(u"<center><img src=\"" +os.path.dirname(os.path.realpath(__file__)) + u"/butler.png\"></center><p>Herzlich willkommen! Das ist das erste Mal, dass das Butler-Addon geladen wurde. Möchtest du die empfohlenen Standard-Einstellungen vornehmen?</p><p>Wenn du auf <b>Ja</b> klickst, wird das Addon:</p><ul><li>Die <a href='http://thomaskahn.de/jura-vorlagen/'>Jura-Vorlagen</a> installieren bzw. auf die neueste Version aktualisieren</li><li>Die <a href=\"" +recommended_plugins_info +u"\">empfohlenen Anki-Addons</a> installieren</li><li>Den Anki-Browser und die allgemeinen Wiederholungsoptionen <a href=\"" +butler_addon_settings_info +u"\">optimal für akademisches Wissen einstellen</a></li></ul><p><u>Wichtig</u>: Am Inhalt deiner Karten und an deinem Lernfortschritt ändert sich dadurch nichts!</p><p>Du kannst natürlich auch auf <b>Nein</b> klicken und nur einzelne (oder gar keine) dieser Einstellungen vornehmen, indem du im Anki-Hauptmenu auf die neue Schaltfläche \"Butler-Addon\" klickst.</p><p>In jedem Fall: Viel Erfolg mit Anki!</p>", title=u"Butler-Addon"):
            configure_settings(silent = True)
            try:
                rec_plugins_stand = http_get(recommended_plugins_stand_remote_21);
            except:
                showInfo(u"Keine Verbindung zum Server möglich!\n\n(Stelle: ask_on_first_start)", title="Butler-Addon")
                save_config(first_start = True)
                return
            if rec_plugins_stand:
                try:
                    recommended_plugins_stand_lokal_time_object = datetime.datetime.strptime(get_recommended_plugins_stand_lokal(), "%Y-%m-%d %H:%M")
                except:
                    recommended_plugins_stand_lokal_time_object = datetime.datetime.strptime(recommended_plugins_stand_lokal, "%Y-%m-%d %H:%M")
                if recommended_plugins_stand_lokal_time_object  < datetime.datetime.strptime(rec_plugins_stand[:16], "%Y-%m-%d %H:%M"):
                    download_recommended_plugins(silent = True)
            if get_jura_vorlagen_version() == -1:
                dl_beta_png()
                showInfo(u"Fast fertig! Jetzt müssen nur noch die Jura-Vorlagen installiert werden. Das kann bis zu 20 sec dauern.<br><br>Danach erscheint ein verwirrendes Textfenster mit lauter HTML-Code. 😅<br><br>Dort einfach auf \"Schließen\" klicken und <b>danach Anki bitte einmal neustarten!</b>", title="Butler-Addon")
                global bk_tagged
                bk_tagged = True
                global bk_tagged2
                bk_tagged2 = True
                save_config()
                download_jura_vorlagen()
            else:
                jv_stand_remote = http_get(jura_vorlagen_stand_remote)
                if jv_stand_remote and get_jura_vorlagen_version() < int(jv_stand_remote.split(" ")[0]):
                    update_jura_vorlagen(False)
                add_basiskarten_tag_to_bk_notes()
                dl_beta_png()
                showInfo(u"Schon fertig mit allem! 👏 Jetzt bitte Anki einmal neustarten!", title="Butler-Addon")
                
def toggle_beta_setting():
    global beta_participate
    if not beta_participate:
        if askUser(u"<center><img src='" +os.path.dirname(os.path.realpath(__file__)) +"/beta.png'></center><br>Möchtest du auf <b>Beta-Versionen</b> hingewiesen werden? Dadurch hast du früher Zugriff auf neue Features, aber Beta-Versionen sind natürlich oft noch nicht völlig stabil.<br><br>Möchtest du am Beta-Programm teilnehmen?", title=u"Butler-Addon - Beta-Programm"): #, sodass schlimmstenfalls Daten verloren gehen können (Deshalb hast du vor der Installation einer Beta-Version auch immer nochmal Gelegenheit, ein manuelles Backup deiner Sammlung zu erstellen.)
            beta_participate = True

            global beta_email            
            antwort = getText(u"Wenn du magst, kannst du hier noch deine E-Mail-Adresse eingeben.<br><br>Diese wird nur dann an meinen Server übertragen, wenn du eine Beta-Version installierst.<br>Das ist extrem hilfreich für mich, weil ich dann genau denjenigen, die eine bestimmte Version<br>ausprobiert haben, falls nötig, noch weitere Fragen dazu stellen kann. (Es besteht natürlich<br>keine Pflicht, darauf zu antworten.)", title=u"Butler-Addon - Beta-Programm", default=beta_email)
            if antwort[1] == 0:
                pass
            else:
                beta_email = antwort[0]
        else:
            beta_participate = False
    else:
        beta_participate = False
        
    global beta_menu_option
    beta_menu_option.setChecked(beta_participate)
    
    global beta_asked
    beta_asked = True
    save_config()

def check_for_butler_addon_update(info_all = None, notify_of_result = True):
    if info_all == None:
      info_all = get_info(notify_of_result = notify_of_result)

    if info_all == None:
      return False
      
    version_remote = info_all['butler_v']
    
    if butler_addon_version_lokal < version_remote:
        if askUser(u"Es gibt eine neue Version des <font color=blue><b>Butler-Addons</b></font>! (Aktuell: " +str(version_remote) +u", deine Version: " +str(butler_addon_version_lokal) +u") Mehr Infos dazu <a href='" +butler_addon_changelog +u"'>hier</a>. Jetzt herunterladen?", title=u"Butler-Addon"):
            try:
                dl_file_to(butler_addon_dl, os.path.realpath(__file__))
            except:
                if notify_of_result:
                    showInfo(u"Keine Verbindung zum Server möglich! (Stelle: check_for_butler_addon_update)")
                return False
    elif notify_of_result:
        showInfo(u"Deine Version des <font color=blue><b>Butler-Addons</b></font> (" +str(butler_addon_version_lokal) +u") ist aktuell!")
    return True
        
def check_for_butler_addon_update_report():
    return check_for_butler_addon_update()

def get_info(notify_of_result = True):

    print("Getting info on latest versions from server (Butler-Addon).")
    
    try:
        info_all = http_get(butler_addon_latest_info)
    except:
        if notify_of_result:
          showInfo(u"Fehler: Keine Verbindung zum Server möglich!\n\n(Stelle: get_info #1; Zeile: {})".format(sys._getframe().f_lineno), title="Butler-Addon")
        return None
    if not info_all:
        if notify_of_result:
          showInfo(u"Leere Antwort vom Server erhalten. Das liegt nicht an dir, da stimmt gerade etwas mit meinem Server nicht. 🚧\n\n(Stelle: get_info #2; Zeile: {})".format(sys._getframe().f_lineno), title="Butler-Addon")
        return None
    try:
      return json.loads(info_all)
    except:
      if notify_of_result:
          showInfo(u"Konnte Antwort vom Server nicht interpretieren. Das liegt nicht an dir, da stimmt gerade etwas mit meinem Server nicht. 🚧\n\n(Stelle: get_info #3; Zeile: {})".format(sys._getframe().f_lineno), title="Butler-Addon")
      return None

def check_all_updates_silent():
    if not first_start:

      second_start = False
      if not beta_asked:
        toggle_beta_setting()
        second_start = True

      #Letzten Stand zu allen Basiskarten, Butler-Addon, empfohlenen Addons und Jura-Vorlagen laden
      info_all = get_info(notify_of_result = False)
      if info_all == None:
        return False

      if butler_addon_automatic_updates:
          check_for_butler_addon_update(info_all = info_all, notify_of_result = False)
      if recommended_plugins_automatic_updates:
          check_for_recommended_plugins_update(info_all = info_all, notify_of_result = False)
      if jura_vorlagen_automatic_updates:
          check_for_jura_vorlagen_update(info_all = info_all, notify_of_result = False)
      if bk_automatic_updates:
          check_for_deck_updates(info_all = info_all, notify_of_result = False)

      global bk_tagged
      global bk_tagged2
      if not bk_tagged or not bk_tagged2:
          add_basiskarten_tag_to_bk_notes()
          bk_tagged  = True
          bk_tagged2 = True
          save_config()

      if not second_start and nachrichten_empfangen:
        check_for_message(info_all = info_all, notify_of_result = False)
        
      print("Update check finished (Butler-Addon).\n")

def check_all_updates_silent_delayed():
    mw.progress.timer(300, check_all_updates_silent, False)
    
def check_all_updates():
    check_for_butler_addon_update()
    check_for_recommended_plugins_update()
    check_for_jura_vorlagen_update()
    check_for_deck_updates() #Auch, wenn kein Update vorliegt, wird eine Antwort gegeben.

def save_config(first_start = False):

    settings = {
        'recommended_plugins_stand_lokal' : recommended_plugins_stand_lokal,
        'first_start' : first_start,
        'butler_addon_automatic_updates' : butler_addon_automatic_updates,
        'bk_automatic_updates' : bk_automatic_updates,
        'jura_vorlagen_automatic_updates' : jura_vorlagen_automatic_updates,
        'recommended_plugins_automatic_updates' : recommended_plugins_automatic_updates,
        'basiskarten_tagged'  : bk_tagged,
        'basiskarten_tagged2' : bk_tagged2,
        'nachrichten_empfangen' : nachrichten_empfangen,
        'nachrichten_stand_lokal' : nachrichten_stand_lokal,
        'beta_asked' : beta_asked,
        'beta_email' : beta_email,
        'beta_participate' : beta_participate,
        'token' : token, 
        'vv_hinweis_anzeigen' : vv_hinweis_anzeigen, 
    }
    try:
      mw.col.set_config('1439937507_butler_addon_config', settings)
      print("Butler-Addon: set_config worked")
    except:
      mw.col.conf['1439937507_butler_addon_config'] = settings
      mw.col.setMod()
      mw.reset()
    
    # pfad = os.path.dirname(os.path.realpath(__file__))
    # if isWin:
        # separator = "\\"
    # else:
        # separator = "/"
    # with open(pfad +separator +'config.json', 'w') as outfile:
        # json.dump(settings, outfile)

def read_config():#Reading config.json file from addon folder (works on Anki 2.0 and 2.1 alike)
    global recommended_plugins_stand_lokal
    global first_start
    global butler_addon_automatic_updates
    global bk_automatic_updates
    global jura_vorlagen_automatic_updates
    global recommended_plugins_automatic_updates
    global bk_tagged
    global bk_tagged2
    global nachrichten_empfangen
    global nachrichten_stand_lokal
    global beta_participate
    global beta_email
    global beta_asked
    global beta_menu_option
    global token
    global vv_hinweis_anzeigen
    
    settings = None
       
    # Load settings from database - new ways
    try:
      settings = mw.col.get_config('1439937507_butler_addon_config')
      print("Butler-Addon: get_config worked")
    except:
    
      # Load settings from database - old way
      if '1439937507_butler_addon_config' in mw.col.conf:
          settings = mw.col.conf['1439937507_butler_addon_config']
          
      # Load settings from file (only if database has no settings)
      else:
          pfad = os.path.dirname(os.path.realpath(__file__))
          try:
              with open(pfad +pfad_trenner +'config.json', 'r') as infile:
                  settings = json.load(infile) #json.loads klappt hier aus irgendeinem Grund nicht.
          except:
              pass

    # showInfo(str(settings))
    if settings != None:
        addons_stand_direkt_aus_plugin_ordner_gelesen = get_recommended_plugins_stand_lokal()
        if addons_stand_direkt_aus_plugin_ordner_gelesen:
            recommended_plugins_stand_lokal = addons_stand_direkt_aus_plugin_ordner_gelesen
        else:
            recommended_plugins_stand_lokal = settings['recommended_plugins_stand_lokal']
        
        try:
            first_start = settings['first_start']
        except:
            pass
        try:
            butler_addon_automatic_updates = settings['butler_addon_automatic_updates']
        except:
            pass
        try:
            bk_automatic_updates = settings['bk_automatic_updates']
        except:
            pass
        try:
            jura_vorlagen_automatic_updates = settings['jura_vorlagen_automatic_updates']
        except:
            pass
        try:
            recommended_plugins_automatic_updates = settings['recommended_plugins_automatic_updates']
        except:
            pass
        try:
            bk_tagged = settings['basiskarten_tagged']
        except:
            pass
        try:
            bk_tagged2 = settings['basiskarten_tagged2']
        except:
            pass
        try:
            nachrichten_empfangen   = settings['nachrichten_empfangen']
        except:
            pass
        try:
            nachrichten_stand_lokal = settings['nachrichten_stand_lokal']
        except:
            pass
        try:
            beta_participate = settings['beta_participate']
        except:
            pass
        try:
            beta_email       = settings['beta_email']
        except:
            pass
        try:
            beta_asked       = settings['beta_asked']
        except:
            pass
        try:
            token = settings['token']
        except:
            pass
        try:
            vv_hinweis_anzeigen = settings['vv_hinweis_anzeigen']
        except:
            pass
        
def datenschutz_info():
    showInfo(u"<center><img src=\"" +os.path.dirname(os.path.realpath(__file__)) + u"""/butler.png\"></center><h3>Butler-Addon: Impressum und Datenschutz</h3>

<p>Verantwortlich für dieses Addon ist: Thomas Kahn, Seelingstr. 34, 14059 Berlin, Deutschland, thomas@basiskarten.de</p>

<p>Wenn das Addon prüft, ob es ein Update oder eine neue Nachricht gibt, verbindet es sich mit einem meiner Server, wodurch zwangsläufig deine IP-Adresse dort bekannt wird.</p>

<p>Bei Karten, die Teil eines Basiskarten-Stapels sind (genauer: bei Karten mit dem Tag "Basiskarten"), wird standardmäßig überprüft, ob andere Basiskarten-User Verbesserungsvorschläge dafür gemacht haben. Falls ja, wird auf der Antwortseite ggf. ein Punkt neben dem Wort "Verbesserungsvorschlag" angezeigt. Um das zu ermöglichen, werden Informationen über die Karte (Basiskarten-ID etc.) sowie ggf. dein persönliches Basiskarten-Token an meinen Server geschickt. Falls du das nicht möchtest, kannst du diese Funktion ausstellen, indem du auf: "Butler-Addon > Basiskarten > Auf Antwortseite auf vorhandene Verbesserungsvorschläge hinweisen" klickst und das Häkchen entfernst.</p>

<p>Deine IP-Adresse wird höchstens gespeichert, um missbräuchliche Zugriffe zu verhindern (z.B. Schutz vor DDOS-Attacken). Eine weitere Verwertung findet nicht statt.</p>

<p>Beta-Programm: Wenn du möchtest, kannst du am Beta-Programm teilnehmen und mir dafür deine E-Mail-Adresse mitteilen. Diese wird nur dann an meinen Server übertragen, wenn du eine Beta-Version installierst. Das ist extrem hilfreich für mich, weil ich dann genau denjenigen, die eine bestimmte Version ausprobiert haben, falls nötig, noch weitere Fragen dazu stellen kann. Es besteht natürlich keine Pflicht, darauf zu antworten, und du kannst dich jederzeit wieder abmelden. Und natürlich wird deine E-Mail-Adresse auch für keine anderen Zwecke verwendet.</p>

<p>Sonstige Daten werden nicht erhoben.</p>""")
        
def butler_addon_toggle_automatic_updates():
    global butler_addon_automatic_updates
    butler_addon_automatic_updates = not butler_addon_automatic_updates
    save_config()
  
menu_added = False
rec_plugins_check_updates_menu_option    = None
juvo_check_updates_menu_option           = None
bk_check_updates_menu_option             = None
bk_show_vv_menu_option                   = None
butler_addon_check_updates_menu_option   = None
beta_menu_option                         = None
messages_addon_check_updates_menu_option = None

def add_basiskarten_tag_to_bk_notes_with_result():
  add_basiskarten_tag_to_bk_notes(show_result = True)

def speedtest():
  start_time = time.time()
  http_get("http://thomaskahn.de/anki-butler-addon/API/get_latest_all.php")
  print("Finished at %s" % (time.time() - start_time))

def add_basiskarten_option_to_menu():
    global menu_added
    if menu_added:
        return
    
    menu_added = True
    
    #global German
    German = True
    
    global rec_plugins_check_updates_menu_option
    global juvo_check_updates_menu_option
    global bk_check_updates_menu_option
    global bk_show_vv_menu_option
    global butler_addon_check_updates_menu_option
    global beta_menu_option
    global messages_addon_check_updates_menu_option
    
    #menu_added += 1
    #if not menu_added > 1:
    try:
        m = mw.menuView
    except:
        mw.menuView = QMenu("B&utler-Addon")
        action = mw.menuBar().insertMenu(
            mw.form.menuTools.menuAction(), mw.menuView)
        m = mw.menuView

    if German:    
        
        #rp1000 = m.addAction(u"Speedtest")
        
        
        rec_plugins = m.addMenu(u"Empfohlene Addons")
        rp1 = rec_plugins.addAction(u"Letzte stabile Version laden")
        rp4 = rec_plugins.addAction(u"Aktuelle Beta-Version laden")
        rp2 = rec_plugins.addAction(u"Jetzt prüfen, ob es eine neue Version gibt")
        rec_plugins_check_updates_menu_option = rec_plugins.addAction(u"Automatisch prüfen, ob es eine neue Version gibt")
        
        ju_vo = m.addMenu(u"Jura-Vorlagen")
        jv1 = ju_vo.addAction(u"Herunterladen und installieren")
        jv3 = ju_vo.addAction(u"Auf letzte stabile Version updaten")
        jv9 = ju_vo.addAction(u"Auf aktuelle Beta-Version updaten")
        jv2 = ju_vo.addAction(u"Jetzt prüfen, ob es eine neue Version gibt")
        juvo_check_updates_menu_option = ju_vo.addAction(u"Automatisch prüfen, ob es eine neue Version gibt")
        #jv4 = ju_vo.addAction(u"Darstellung ändern")
        #jv5 = ju_vo.addAction(u"Darstellung zurücksetzen")
        jv6 = ju_vo.addAction(u"Schriftgröße ändern")
        jv7 = ju_vo.addAction(u"Farbe von Links und Normen ändern")
        #jv7 = ju_vo.addAction(u"Festlegen, für welche Karten geprüft wird, ob es Verbesserungsvorschläge gibt")
    
        bk = m.addMenu(u"Basiskarten")
        bk_check_updates_menu_option = bk.addAction(u"Automatisch prüfen, ob es Updates für deine Stapel gibt")
        bk1 = bk.addAction(u"Jetzt prüfen, ob es Updates für deine Stapel gibt")
        #bk4 = bk.addAction(u"Basiskarten taggen (wichtig für Verbesserungsvorschlag-Anzeige + Updates)")
        bk5 = bk.addAction(u"Token eingeben oder ändern (wichtig für Updates)")
        bk_show_vv_menu_option = bk.addAction(u"Auf Antwortseite auf vorhandene Verbesserungsvorschläge hinweisen")
        #bk6 = bk.addAction(u"Basiskarten-Stapelbeschreibungen wiederherstellen (wichtig für Updates) [Ansatz 1]")
        #bk7 = bk.addAction(u"Ausgewählten Stapel als Basiskarten-Stapel deklarieren [Ansatz 2]")
        
        nuetzlich = m.addMenu(u"Sonstiges")
        n1 = nuetzlich.addAction(u"Anki optimal einstellen")
        h  = nuetzlich.addAction(u"Alte Tabellenformatierung entfernen")
        n2 = nuetzlich.addAction(u"Medien-Ordner öffnen")
        n3 = nuetzlich.addAction(u"Addon-Ordner öffnen")
        n4 = nuetzlich.addAction(u"Backup-Ordner öffnen")
        n5 = nuetzlich.addAction(u"Update-Log-Ordner öffnen")
        msg2 = nuetzlich.addAction(u"Letzte Nachricht abrufen")
        
        #b = m.addAction(u"Jetzt auf Updates prüfen")
        #f = m.addAction(u"Dump Jura-Vorlagen-models to \"jura_vorlagen.txt\"")
        #i = m.addAction(u"HTML aus Feldnamen der Jura-Vorlagen entfernen")
        t   = m.addAction(u"Jetzt prüfen, ob es ein Update für das Butler-Addon gibt")
        butler_addon_check_updates_menu_option = m.addAction(u"Automatisch prüfen, ob es Updates für das Butler-Addon gibt")
        beta_menu_option = m.addAction(u"An Beta-Programm teilnehmen")
        messages_addon_check_updates_menu_option = m.addAction(u"Nachrichten zu Anki, Basiskarten usw. erhalten")
        ds  = m.addAction(u"Impressum und Datenschutz")
        m.addSeparator()
    else:
        pass
    n1.triggered.connect(configure_settings)
    n2.triggered.connect(open_media_folder)
    n3.triggered.connect(open_addon_folder)
    n4.triggered.connect(open_backup_folder)
    n5.triggered.connect(open_update_log_folder)
    #b.triggered.connect(check_all_updates)
    rp1.triggered.connect(download_recommended_plugins_ask_first)
    #rp1000.triggered.connect(speedtest)
    rp4.triggered.connect(download_recommended_plugins_beta_ask_first)
    rp2.triggered.connect(check_for_recommended_plugins_update_report)
    rec_plugins_check_updates_menu_option.setCheckable(True)
    rec_plugins_check_updates_menu_option.triggered.connect(recommended_plugins_toggle_automatic_updates)
    
    h.triggered.connect(remove_formatting_from_tables_in_cards)
    #f.triggered.connect(dump_jura_vorlagen_models)
    jv1.triggered.connect(download_jura_vorlagen_ask_first)
    jv2.triggered.connect(check_for_jura_vorlagen_update_report)
    jv3.triggered.connect(update_jura_vorlagen_ask_first)
    jv9.triggered.connect(update_jura_vorlagen_beta_ask_first)
    jv6.triggered.connect(change_font_size_of_jura_vorlagen)
    jv7.triggered.connect(change_color_of_links)
    juvo_check_updates_menu_option.setCheckable(True)
    juvo_check_updates_menu_option.triggered.connect(jura_vorlagen_toggle_automatic_updates)    
    
    bk1.triggered.connect(check_for_deck_updates_report)
    bk_check_updates_menu_option.setCheckable(True)
    bk_check_updates_menu_option.triggered.connect(bk_toggle_automatic_updates)
    bk_show_vv_menu_option.setCheckable(True)
    bk_show_vv_menu_option.triggered.connect(bk_toggle_vv_anzeige)
    #bk3.setCheckable(True)
    #bk3.setChecked(bk_auf_verbesserungsvorschlaege_pruefen)
    #bk4.triggered.connect(add_basiskarten_tag_to_bk_notes_with_result)
    bk5.triggered.connect(set_token)  
    
    ds.triggered.connect(datenschutz_info)
    t.triggered.connect(check_for_butler_addon_update_report)
    butler_addon_check_updates_menu_option.setCheckable(True)
    butler_addon_check_updates_menu_option.triggered.connect(butler_addon_toggle_automatic_updates)
    messages_addon_check_updates_menu_option.setCheckable(True)
    messages_addon_check_updates_menu_option.triggered.connect(nachrichten_toggle_automatic_updates)
    msg2.triggered.connect(
        lambda a : check_for_message()) #Wenn ich das nicht so mache, wird das erste Argument (egal, wie es heißt) immer auf False gesetzt.
    beta_menu_option.setCheckable(True)
    
    beta_menu_option.triggered.connect(toggle_beta_setting)
    
def update_menu_options():
    global rec_plugins_check_updates_menu_option
    global juvo_check_updates_menu_option
    global bk_check_updates_menu_option
    global bk_show_vv_menu_option
    global butler_addon_check_updates_menu_option
    global beta_menu_option
    global messages_addon_check_updates_menu_option
    
    rec_plugins_check_updates_menu_option.setChecked(recommended_plugins_automatic_updates)
    juvo_check_updates_menu_option.setChecked(jura_vorlagen_automatic_updates)
    bk_check_updates_menu_option.setChecked(bk_automatic_updates)
    bk_show_vv_menu_option.setChecked(vv_hinweis_anzeigen)
    butler_addon_check_updates_menu_option.setChecked(butler_addon_automatic_updates)
    beta_menu_option.setChecked(beta_participate)
    messages_addon_check_updates_menu_option.setChecked(nachrichten_empfangen)
    
def start_update_timer():
    mw.progress.timer(update_interval * 1000 * 60, check_all_updates_silent, True)
 
def ask_on_first_start_timer():
    save_config()
    mw.progress.timer(500, ask_on_first_start, False)

# Notiz in Zwischenablage kopieren

from PyQt5.QtWidgets import QAction
from aqt.qt import QKeySequence
from aqt.browser import Browser
from anki.utils import ids2str

def get_basiskarten_id(note):
  try:
    m = re.search('\[basiskarten_id:\s+(\d+)\]', note.fields[len(note.fields)-1])
    return m.group(1)
  except:
    return None
    
def get_basiskarten_update_version(note):
  try:
    m = re.search('\[basiskarten_update:\s+(\d+)\]', note.fields[len(note.fields)-1])
    return int(m.group(1))
  except:
    return 0

def copy_selected_note_to_clipboard(notes):
#Struktur: [ [nid, mid, [feld1_name, feld2_name...], [feld1_inhalt, feld2_inhalt...] ], ... ]
    if notes != []:
        note = mw.col.getNote(notes[0])
        #showInfo(json.dumps([notes[0], note.fields]))
        field_names = []
        for fn in note._fmap:
            field_names.append(fn)
        for_clipboard = json.dumps( {
          'nid' : get_basiskarten_id(note),
          'mid' : note.mid,
          'feld_namen' : field_names,
          'feld_inhalte' : note.fields
        } )
        if isMac:
            import subprocess
            process = subprocess.Popen('pbcopy', env={'LANG': 'en_US.UTF-8'}, stdin=subprocess.PIPE)
            process.communicate(for_clipboard.encode('utf-8'))
        else:
            QApplication.clipboard().setText(for_clipboard)
        showInfo("Notiz {} in die Zwischenablage kopiert.".format(get_basiskarten_id(note)))
        return
        
def show_discussion_on_bk_portal(notes):
    if notes != []:
        last_nid = 0
        for n in notes:
          note = mw.col.getNote(n)
          if not "Basiskarten" in note.tags:
            showInfo("Dies ist keine Basiskarte.")
            return
          if n == last_nid:
              continue              
          last_nid = n            
          open_file("https://basiskarten.anvil.app/#?nid={}".format(get_basiskarten_id(note)))

def setupBrowserMenu(browser):
    
    c = QAction("Diskussion im Basiskarten-Portal anzeigen", browser)
    c.setShortcut(QKeySequence("Ctrl+#"))
    c.triggered.connect(lambda: show_discussion_on_bk_portal(browser.selectedNotes()))
    browser.form.menu_Notes.addSeparator()
    browser.form.menu_Notes.addAction(c)
    
    a = QAction("In Zwischenablage kopieren", browser)
    a.setShortcut(QKeySequence("Ctrl+Alt+#"))
    a.triggered.connect(lambda: copy_selected_note_to_clipboard(browser.selectedNotes()))
    #showInfo(str(dir(browser.form)))
                                          
    browser.form.menu_Notes.addAction(a)
    #showInfo("hi2")
    
    browser.form.menu_Notes.addSeparator()

ran_once_after_opening_profile = True
cant_wait_for_sync = False

def profile_loaded_now_do_this():
    # showInfo(str(first_start))
    print("profile_loaded_now_do_this()")
    global ran_once_after_opening_profile
    ran_once_after_opening_profile = False
    
    #Falls Sync aktiviert, diese zuerst abwarten. Falls nicht, sofort starten.
    if not mw.pm.profile.get("autoSync") or mw.pm.profile.get("syncUser") in (None, "") or mw.pm.profile.get("syncKey") in (None, ""):
      butler_start()
    elif not cant_wait_for_sync:
      mw.progress.timer(3000, butler_start, False)

def butler_start():
    global ran_once_after_opening_profile
    if ran_once_after_opening_profile:
      return
    ran_once_after_opening_profile = True
    
    print("butler_start()")
    reset_values()
    read_config()
    save_config()
    settings_datei_fuer_jv_speichern()
    datei_loeschen("_jura_vorlagen.css")
    datei_loeschen("jura_vorlagen.apkg")
    update_menu_options()
    #global first_start
    # showInfo(str(first_start))
    if first_start:
        #showInfo("hi")
        addHook("profileLoaded", ask_on_first_start_timer)
        addHook("profileLoaded", start_update_timer)
    else:
        if butler_addon_automatic_updates or bk_automatic_updates or recommended_plugins_automatic_updates or jura_vorlagen_automatic_updates or nachrichten_empfangen:
            addHook("profileLoaded", check_all_updates_silent_delayed)
            addHook("profileLoaded", start_update_timer)



######################################
#
#   NACH DEM START
#
######################################

#showInfo("loaded")

add_basiskarten_option_to_menu()

addHook("browser.setupMenus", setupBrowserMenu)
addHook("profileLoaded", profile_loaded_now_do_this)
try:
  gui_hooks.sync_did_finish.append(butler_start)
  cant_wait_for_sync = True
except:
  cant_wait_for_sync = False